from polly import helpers
from polly.errors import (
    InvalidCohortOperationException,
    InvalidParameterException,
    InvalidRepoException,
    CohortEditException,
    EmptyCohortException,
    InvalidCohortAddition,
    InvalidCohortMergeOperation,
    InvalidPathException,
)
import os
import pandas as pd
from deprecated import deprecated
from polly.bridge_cohort import CohortRepoClass
from polly.core_cohort import CohortFileStandard
from polly import constants as const
from polly.help import example
from polly.auth import Polly
from polly.tracking import Track


class Cohort:
    """
    The Cohort class contains functions which can be used to create cohorts, add or remove samples, \
    merge metadata and data-matrix of samples/datasets in a cohort and edit or delete a cohort.
    Args:
        token (str): Authentication token from polly
    Usage:
        from polly.cohort import Cohort

        cohort = Cohort(token)
    """

    _cohort_info = None
    example = classmethod(example)

    def __init__(self, token=None, env="", default_env="polly") -> None:
        # check if COMPUTE_ENV_VARIABLE present or not
        # if COMPUTE_ENV_VARIABLE, give priority
        env = helpers.get_platform_value_from_env(
            const.COMPUTE_ENV_VARIABLE, default_env, env
        )
        self.session = Polly.get_session(token, env=env)
        if self._cohort_info is None:
            self._cohort_info = helpers.get_cohort_constants()
        self._cohort_path = None
        self._cohort_details = None
        self.bridge_obj = CohortRepoClass(self.session)
        self.core_obj = CohortFileStandard()

    @Track.track_decorator
    def create_cohort(
        self,
        local_path: str,
        cohort_name: str,
        description: str,
        repo_key=None,
        dataset_id=None,
        sample_id=None,
    ) -> None:
        """
        This function is used to create a cohort. After making Cohort Object you can create cohort.

        Args:
              local_path (str): local path to instantiate the cohort.
              cohort_name (str): identifier name for the cohort.
              description (str): description about the cohort.
              repo_key (str, optional): repo_key(repo_name/repo_id) for the omixatlas \
                from where datasets or samples is to be added.
              dataset_id (list/str, optional): dataset_ids(list,in case of repositories where one dataset has 1 \
                sample) or a dataset_id(str,in case of in case of repository where 1 dataset has many samples)
              sample_id (list, optional): list of samples to be added in cohort, \
                applicable only in case of in case of repository where 1 dataset has many samples.

        Returns:
              A message will be displayed on the status of the operation.

        Raises:
              InvalidParameterException: Empty or Invalid Parameters
              InvalidCohortNameException: The cohort_name does not represent a valid cohort name.
              InvalidPathException: Provided path does not represent a file or a directory.
        """
        self._cohort_path = self.bridge_obj.create_cohort(
            local_path, cohort_name, description
        )
        self._cohort_details = self.bridge_obj._cohort_details
        if repo_key:
            self.add_to_cohort(
                repo_key=repo_key,
                dataset_id=dataset_id,
                sample_id=sample_id,
            )

    @Track.track_decorator
    def add_to_cohort(self, repo_key: str, dataset_id=None, sample_id=None) -> None:
        """
        This function is used to add datasets or samples to a cohort.

        Args:
              repo_key (str): repo_key(repo_name OR repo_id) for the omixatlas where datasets or samples belong.
              dataset_id (list/str): dataset_ids(list,in case of repositories where one dataset has 1 \
                sample) or a dataset_id(str,in case of in case of repository where 1 dataset has many samples)
              sample_id (list, optional): list of samples to be added in cohort, \
                applicable only in case of in case of repository where 1 dataset has many samples.

        Returns:
              A message will be displayed on the status of the operation.

            Raises:
              InvalidParameterException: Empty or Invalid Parameters.
              InvalidCohortOperationException: This operation is not valid as no cohort has been instantiated.
        """
        self.bridge_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if not ((repo_key and dataset_id and sample_id) or (repo_key and dataset_id)):
            raise InvalidParameterException("repo_key dataset_id sample_id")
        repo_name = self.bridge_obj._return_omixatlas_name(repo_key)
        if repo_name not in self._cohort_info:
            raise InvalidRepoException(repo_name)
        cohort_entity_type = self._assign_entity_type(repo_name, dataset_id)
        if "entity_type" in self._cohort_details:
            existing_entity_type = self._cohort_details.get("entity_type")
            if existing_entity_type != cohort_entity_type:
                raise InvalidCohortAddition
        self.bridge_obj._cohort_details = self._cohort_details
        if cohort_entity_type == "dataset":
            self.bridge_obj.add_to_cohort_single(
                self._cohort_path,
                repo_name=repo_name,
                entity_id=dataset_id,
                entity_type=cohort_entity_type,
            )
        else:
            self.bridge_obj.add_to_cohort_multiple(
                self._cohort_path,
                repo_name=repo_name,
                dataset_id=dataset_id,
                sample_id=sample_id,
                entity_type=cohort_entity_type,
            )
        self._cohort_details = self.bridge_obj._cohort_details

    @Track.track_decorator
    def remove_from_cohort(self, dataset_id=None, sample_id=[]) -> None:
        """
        This function is used for removing datasets or samples from a cohort.

        Args:
              dataset_id (list/str): dataset_ids(list,in case of repositories where one dataset has 1 \
                sample) or a dataset_id(str,in case of in case of repository where 1 dataset has many samples)
              sample_id (list, optional): list of samples to be added in cohort, \
                applicable only in case of in case of repository where 1 dataset has many samples.

        Returns:
              A message will be displayed on the status of the operation.

        Raises:
              InvalidParameterException: Empty or Invalid Parameters
              InvalidCohortOperationException: This operation is not valid as no cohort has been instantiated.
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        entity_type = self._cohort_details.get("entity_type")
        if entity_type == "dataset":
            self.core_obj.remove_single_from_cohort(self._cohort_path, dataset_id)
        else:
            self.core_obj.remove_multiple_from_cohort(
                self._cohort_path, dataset_id, sample_id
            )
        self._cohort_details = self.core_obj._cohort_details

    @Track.track_decorator
    def merge_data(self, data_level: str):
        """
        Function to merge metadata (dataset,sample and feature level) or data-matrix of all the samples/datasets in the cohort.

        Args:
                data_level (str): identifier to specify the data to be merged - "dataset", "sample", "feature" or "data_matrix"

        Returns:
                A pandas dataframe containing the merged data which is ready for analysis
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if not (data_level and isinstance(data_level, str)):
            raise InvalidParameterException("data_level")
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        if not (data_level and isinstance(data_level, str)):
            raise InvalidParameterException("data_level")
        if data_level == "sample":
            sample_df = self._merge_sample_metadata()
            return sample_df
        if data_level == "dataset":
            dataset_df = self._merge_dataset_metadata()
            return dataset_df
        if data_level == "data_matrix":
            datamatrix_df = self._merge_data_matrix()
            return datamatrix_df
        if data_level == "feature":
            feature_df = self._merge_feature_metadata()
            return feature_df
        raise InvalidCohortMergeOperation

    @deprecated(reason="use function merge_data")
    def _merge_sample_metadata(self) -> pd.DataFrame:
        """
        Function to merge the sample level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.

        :meta private:
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        entity_type = self._cohort_details.get("entity_type")
        if entity_type == "sample":
            return self.core_obj.merge_multiple_sample_metadata(self._cohort_path)
        else:
            return self.core_obj.merge_single_sample_metadata(self._cohort_path)

    @Track.track_decorator
    def is_valid(self) -> bool:
        """
        This function is used to check the validity of a cohort.

        Returns:
                A boolean result based on the validity of the cohort.

        Raises:
                InvalidPathException: Cohort path does not represent a file or a directory.
                InvalidCohortOperationException: This operation is not valid as no cohort has been instantiated.
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        return self.core_obj.is_valid(self._cohort_path)

    @deprecated(reason="use function merge_data")
    def _merge_feature_metadata(self) -> pd.DataFrame:
        """
        Function to merge the feature level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.

        :meta private:
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        return self.core_obj.merge_feature_metadata(self._cohort_path)

    @deprecated(reason="use function merge_data")
    def _merge_dataset_metadata(self) -> pd.DataFrame:
        """
        Function to merge the dataset level metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.

        :meta private:
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        return self.core_obj.merge_dataset_metadata(self._cohort_path)

    @deprecated(reason="use function merge_data")
    def _merge_data_matrix(self) -> pd.DataFrame:
        """
        Function to merge the data matrix metadata from all the gct files in a cohort.
        Returns:
            | A pandas dataframe containing the merged metadata for analysis.

        :meta private:
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        entity_type = self._cohort_details.get("entity_type")
        if entity_type == "sample":
            return self.core_obj.merge_multiple_data_matrix(self._cohort_path)
        else:
            return self.core_obj.merge_single_data_matrix(self._cohort_path)

    @Track.track_decorator
    def load_cohort(self, local_path: str):
        """
        Function to load an existing cohort into an object.
        Once loaded, the functions described in the documentation can be used for the object where the cohort is loaded.

        Args:
                    local_path (str): local path of the cohort.

        Returns:
                    A confirmation message on instantiation of the cohort.
        Raises:
                    InvalidPathException: This path does not represent a file or a directory.
                    InvalidCohortPathException: This path does not represent a Cohort.

        """
        self.core_obj.load_cohort(local_path)
        self._cohort_path = local_path
        self._cohort_details = self.core_obj._cohort_details

    @Track.track_decorator
    def edit_cohort(self, new_cohort_name=None, new_description=None):
        """
        This function is used to edit the cohort level metadata such as cohort name and description.
        Atleast one of the argument should be present.
        Args:
                    new_cohort_name (str): new identifier name for the cohort.
                    new_description (str): new description about the cohort.

            Returns:
                    A confirmation message on updation of cohort.

            Raises:
                    InvalidCohortOperationException: This operation is not valid as no cohort has been instantiated.
                    CohortEditException: No parameter specified for editing in cohort
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if new_cohort_name is None and new_description is None:
            raise CohortEditException
        new_path = self.core_obj.edit_cohort(
            new_cohort_name, new_description, self._cohort_path
        )
        if new_path:
            self._cohort_path = new_path
        self._cohort_details = self.core_obj._cohort_details

    @Track.track_decorator
    def summarize_cohort(self):
        """
        Function to return cohort level metadata and dataframe with datasets or samples added in the cohort.

        Returns:
              A tuple with the first value as cohort metadata information
              (name, description and number of dataset(s) or sample(s) in the cohort) and the second value
              as dataframe containing the source, dataset_id/sample_id and data type available in the cohort.

            Raises:
              InvalidCohortOperationException: This operation is not valid as no cohort has been instantiated.
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        return self.core_obj.summarize_cohort(self._cohort_path)

    @Track.track_decorator
    def delete_cohort(self) -> None:
        """
        This function is used to delete a cohort.
        Returns:
            A confirmation message on deletion of cohort
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        self.core_obj.delete_cohort(self._cohort_path)
        self.bridge_obj._cohort_details = self.core_obj._cohort_details
        self._cohort_details = self.core_obj._cohort_details
        self._cohort_path = None

    def _assign_entity_type(self, repo_name: str, dataset_id: str) -> str:
        """
        Function to return entity type
        Input: repo_name(str) and dataset_id(str or list)
        Returns: Entity_type('dataset' for single mapped repositories,\
            'sample' for multiple mapped repositories.)
        """
        for repo, dict in self._cohort_info.items():
            if repo_name == repo:
                if dict["file_structure"] == "single":
                    cohort_entity_type = "dataset"
                elif dict["file_structure"] == "multiple":
                    cohort_entity_type = "sample"
                else:
                    cohort_entity_type = self.bridge_obj._get_entity_type(
                        repo_name, dataset_id, self._cohort_info
                    )
        return cohort_entity_type

    @Track.track_decorator
    def create_merged_gct(self, file_path: str, file_name="") -> None:
        """
          This function is used to merge all the gct files in a cohort into a single gct file.

        Args:
                    file_path (str): the system path where the gct file is to be written.
                    file_name (str, optional): Identifier for the merged file name, cohort name will be used by default.
        """
        self.core_obj._cohort_details = self._cohort_details
        if self._cohort_details is None:
            raise InvalidCohortOperationException
        if "entity_type" not in self._cohort_details:
            raise EmptyCohortException
        if not isinstance(file_name, str):
            raise InvalidParameterException("file_name")
        isExists = os.path.exists(file_path)
        if not isExists:
            raise InvalidPathException
        if file_name:
            if not isinstance(file_name, str):
                raise InvalidParameterException("file_name")
            file_name = f"{file_name}.gct"
        else:
            # Fetching cohort name to be used as the file name
            cohort_info = self.summarize_cohort()
            cohort_name = cohort_info[0].get("cohort_name")
            file_name = f"{cohort_name}.gct"
        entity_type = self._cohort_details.get("entity_type")
        self.core_obj.merge_gcts(entity_type, self._cohort_path, file_name, file_path)



from abc import ABC, abstractmethod
from typing import Generic

from aws_lambda_powertools.event_handler.api_gateway import Response
from aws_lambda_powertools.event_handler.types import EventHandlerInstance
from aws_lambda_powertools.shared.types import Protocol


class NextMiddleware(Protocol):
    def __call__(self, app: EventHandlerInstance) -> Response:
        """Protocol for callback regardless of next_middleware(app), get_response(app) etc"""
        ...

    def __name__(self) -> str:  # noqa A003
        """Protocol for name of the Middleware"""
        ...


class BaseMiddlewareHandler(Generic[EventHandlerInstance], ABC):
    """Base implementation for Middlewares to run code before and after in a chain.


    This is the middleware handler function where middleware logic is implemented.
    The next middleware handler is represented by `next_middleware`, returning a Response object.

    Examples
    --------

    **Correlation ID Middleware**

    ```python
    import requests

    from aws_lambda_powertools import Logger
    from aws_lambda_powertools.event_handler import APIGatewayRestResolver, Response
    from aws_lambda_powertools.event_handler.middlewares import BaseMiddlewareHandler, NextMiddleware

    app = APIGatewayRestResolver()
    logger = Logger()


    class CorrelationIdMiddleware(BaseMiddlewareHandler):
        def __init__(self, header: str):
            super().__init__()
            self.header = header

        def handler(self, app: APIGatewayRestResolver, next_middleware: NextMiddleware) -> Response:
            # BEFORE logic
            request_id = app.current_event.request_context.request_id
            correlation_id = app.current_event.get_header_value(
                name=self.header,
                default_value=request_id,
            )

            # Call next middleware or route handler ('/todos')
            response = next_middleware(app)

            # AFTER logic
            response.headers[self.header] = correlation_id

            return response


    @app.get("/todos", middlewares=[CorrelationIdMiddleware(header="x-correlation-id")])
    def get_todos():
        todos: requests.Response = requests.get("https://jsonplaceholder.typicode.com/todos")
        todos.raise_for_status()

        # for brevity, we'll limit to the first 10 only
        return {"todos": todos.json()[:10]}


    @logger.inject_lambda_context
    def lambda_handler(event, context):
        return app.resolve(event, context)

    ```

    """

    @abstractmethod
    def handler(self, app: EventHandlerInstance, next_middleware: NextMiddleware) -> Response:
        """
        The Middleware Handler

        Parameters
        ----------
        app: EventHandlerInstance
            An instance of an Event Handler that implements ApiGatewayResolver
        next_middleware: NextMiddleware
            The next middleware handler in the chain

        Returns
        -------
        Response
            The response from the next middleware handler in the chain

        """
        raise NotImplementedError()

    @property
    def __name__(self) -> str:  # noqa A003
        return str(self.__class__.__name__)

    def __call__(self, app: EventHandlerInstance, next_middleware: NextMiddleware) -> Response:
        """
        The Middleware handler function.

        Parameters
        ----------
        app: ApiGatewayResolver
            An instance of an Event Handler that implements ApiGatewayResolver
        next_middleware: NextMiddleware
            The next middleware handler in the chain

        Returns
        -------
        Response
            The response from the next middleware handler in the chain
        """
        return self.handler(app, next_middleware)



"""Standalone functions to serialize/deserialize common data structures"""
import base64
import json
from typing import Any, Callable


def base64_encode(data: str) -> str:
    """Encode a string and returns Base64-encoded encoded value.

    Parameters
    ----------
    data: str
        The string to encode.

    Returns
    -------
    str
        The Base64-encoded encoded value.
    """
    return base64.b64encode(data.encode()).decode("utf-8")


def base64_decode(data: str) -> str:
    """Decodes a Base64-encoded string and returns the decoded value.

    Parameters
    ----------
    data: str
        The Base64-encoded string to decode.

    Returns
    -------
    str
        The decoded string value.
    """
    return base64.b64decode(data).decode("utf-8")


def base64_from_str(data: str) -> str:
    """Encode str as base64 string"""
    return base64.b64encode(data.encode()).decode("utf-8")


def base64_from_json(data: Any, json_serializer: Callable[..., str] = json.dumps) -> str:
    """Encode JSON serializable data as base64 string

    Parameters
    ----------
    data: Any
        JSON serializable (dict, list, boolean, etc.)
    json_serializer: Callable
        function to serialize `obj` to a JSON formatted `str`, by default json.dumps

    Returns
    -------
    str:
        JSON string as base64 string
    """
    return base64_from_str(data=json_serializer(data))



from __future__ import annotations

from typing import Any, Callable, Dict

from aws_lambda_powertools.logging.formatter import LambdaPowertoolsFormatter
from aws_lambda_powertools.logging.types import LogRecord


class DatadogLogFormatter(LambdaPowertoolsFormatter):
    def __init__(
        self,
        json_serializer: Callable[[LogRecord], str] | None = None,
        json_deserializer: Callable[[Dict | str | bool | int | float], str] | None = None,
        json_default: Callable[[Any], Any] | None = None,
        datefmt: str | None = None,
        use_datetime_directive: bool = False,
        log_record_order: list[str] | None = None,
        utc: bool = False,
        use_rfc3339: bool = True,  # NOTE: The only change from our base formatter
        **kwargs,
    ):
        """Datadog formatter to comply with Datadog log parsing

        Changes compared to the default Logger Formatter:

        - timestamp format to use RFC3339 e.g., "2023-05-01T15:34:26.841+0200"


        Parameters
        ----------
        log_record_order : list[str] | None, optional
            _description_, by default None

        Parameters
        ----------
        json_serializer : Callable, optional
            function to serialize `obj` to a JSON formatted `str`, by default json.dumps
        json_deserializer : Callable, optional
            function to deserialize `str`, `bytes`, bytearray` containing a JSON document to a Python `obj`,
            by default json.loads
        json_default : Callable, optional
            function to coerce unserializable values, by default str

            Only used when no custom JSON encoder is set

        datefmt : str, optional
            String directives (strftime) to format log timestamp.

            See https://docs.python.org/3/library/time.html#time.strftime or
        use_datetime_directive: str, optional
            Interpret `datefmt` as a format string for `datetime.datetime.strftime`, rather than
            `time.strftime` - Only useful when used alongside `datefmt`.

            See https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior . This
            also supports a custom %F directive for milliseconds.

        log_record_order : list, optional
            set order of log keys when logging, by default ["level", "location", "message", "timestamp"]

        utc : bool, optional
            set logging timestamp to UTC, by default False to continue to use local time as per stdlib
        use_rfc3339: bool, optional
            Whether to use a popular dateformat that complies with both RFC3339 and ISO8601.
            e.g., 2022-10-27T16:27:43.738+02:00.
        kwargs
            Key-value to persist in all log messages
        """
        super().__init__(
            json_serializer=json_serializer,
            json_deserializer=json_deserializer,
            json_default=json_default,
            datefmt=datefmt,
            use_datetime_directive=use_datetime_directive,
            log_record_order=log_record_order,
            utc=utc,
            use_rfc3339=use_rfc3339,
            **kwargs,
        )



from functools import wraps
from mixpanel import Mixpanel
import time
from polly.__init__ import __version__
from polly.constants import MIXPANEL_KEY


class Track:
    """
    This class is used for tracking polly-python services using mixpanel.
    This class will have a unidirectional association relationship with the class which wants to use the tracking service.
    The track function sends the logs to mixpanel.
    For using this feature, make an object in the __init__ function for the class this feature is to be used,
    then use the track function with that object.
    """

    def track_decorator(function):
        @wraps(function)
        def wrapper_function(*args, **kwargs):
            execution_flag = False
            try:
                result = function(*args, **kwargs)
                execution_flag = True
            except Exception as e:
                returned_err = e
            obj = args[0].session
            tracking_env = obj.env
            email_id = obj.user_details.get("email")
            properties = {"email-id": email_id}
            # checking if args is not empty(contains self object apart from arguments)
            if len(args) > 1:
                args_list = []
                for index in range(1, len(args)):
                    args_list.append(args[index])
                properties["arguments"] = args_list
            # checking if kwargs exist
            if kwargs:
                for key, value in kwargs.items():
                    properties[key] = value
            # tracking the execution status (True/False) of the function
            properties["execution_status"] = execution_flag
            # tracking the env (dev/test/prod) in which this function was called
            properties["tracking_env"] = tracking_env
            # current polly py version
            properties["pollypy_version"] = __version__

            # current timestamp
            # reference link
            # https://stackoverflow.com/questions/16755394/
            # what-is-the-easiest-way-to-get-current-gmt-time-in-unix-timestamp-format
            properties["current_timestamp"] = time.time()

            mp = Mixpanel(MIXPANEL_KEY)
            mp.track(email_id, function.__name__, properties)
            if execution_flag:
                if result is not None:
                    return result
            else:
                raise returned_err

        return wrapper_function



import stat
import hashlib
import itertools
import logging
import math
from collections import deque
from concurrent.futures import ThreadPoolExecutor
from threading import Lock
import os
from os import PathLike
import pathlib
from typing import Optional, Union
from urllib.parse import (
    parse_qs,
    quote,
    unquote,
    urlencode,
    urlparse,
    urlunparse,
)

from urllib.request import pathname2url, url2pathname
from tqdm import tqdm
from boto3.s3.transfer import TransferConfig
from s3transfer.utils import (
    ChunksizeAdjuster,
    OSUtils,
)
from botocore.exceptions import (
    ClientError,
)
from botocore.client import BaseClient

s3_transfer_config = TransferConfig()

logger = logging.getLogger(__name__)

MAX_CONCURRENCY = 10

# When uploading files at least this size, compare the ETags first and skip the upload
# if they're equal; copy the remote file onto itself if the metadata changes.
UPLOAD_ETAG_OPTIMIZATION_THRESHOLD = 1024


class URLParseError(ValueError):
    pass


def read_bytes(client: BaseClient, s3_uri: str) -> bytes:
    pk = _PhysicalKey.from_pathlike(s3_uri)
    response = client.get_object(Bucket=pk.bucket, Key=pk.path, VersionId=pk.version_id)
    return response["Body"].read()


def copy_file(
    src_uri: Union[str, PathLike],
    dest_uri: Union[str, PathLike],
    *,
    s3_client: BaseClient,
    disable_tqdm: bool = False,
    tqdm_desc=None,
) -> str:
    """
    Can be used to copy file from
    - local to s3
    - s3 to local
    - s3 to s3

    Implements the following
    - Multithreading
    - Multi-part uploads and downloads depending on size
    - Returns full s3 uri in response, including versionId (unlike `boto3.upload_file`)
    - Skips upload if local file has same md5sum as remote file
    - Tqdm progress bar

    Returns full destination URI
    E.g. 'file:///home/me/file.txt', 's3://path/to/file.txt?versionId=abcd'
    """
    transfer = _FileTransfer(src_uri, dest_uri, s3_client, disable_tqdm, tqdm_desc)
    return transfer.start()


def _fix_url(url):
    """Convert non-URL paths to file:// URLs"""
    # If it has a scheme, we assume it's a URL.
    if not url:
        raise ValueError("Empty URL")

    url = str(url)

    parsed = urlparse(url)
    if parsed.scheme and not os.path.splitdrive(url)[0]:
        return url

    fixed_url = pathlib.Path(url).expanduser().resolve().absolute().as_uri()

    # pathlib likes to remove trailing slashes, so add it back if needed.
    if url[-1:] in (os.sep, os.altsep) and not fixed_url.endswith("/"):
        fixed_url += "/"

    return fixed_url


class _PhysicalKey:
    __slots__ = ("bucket", "path", "version_id")

    def __init__(self, bucket: Optional[str], path: str, version_id: Optional[str]):
        """
        For internal use only; call from_path or from_url instead.
        """
        if bucket is not None and not isinstance(bucket, str):
            raise ValueError("Bucket must be None or a string")
        if not isinstance(path, str):
            raise ValueError("Path must be a string")
        if version_id is not None and not isinstance(version_id, str):
            raise ValueError("Version ID must be None or a string")

        if bucket is None:
            if path is None:
                raise ValueError("Local keys must have a path")
            if version_id is not None:
                raise ValueError("Local keys cannot have a version ID")
            if os.name == "nt":
                if "\\" in path:
                    raise ValueError("Paths must use / as a separator")
        else:
            if path.startswith("/"):
                raise ValueError("S3 paths must not start with '/'")

        self.bucket = bucket
        self.path = path
        self.version_id = version_id

    @classmethod
    def from_pathlike(cls, path: Union[str, PathLike]):
        return cls.from_url(_fix_url(path))

    @classmethod
    def from_url(cls, url):
        parsed = urlparse(url)

        if parsed.scheme == "s3":
            if not parsed.netloc:
                raise URLParseError("Missing bucket")
            bucket = parsed.netloc
            assert not parsed.path or parsed.path.startswith("/")
            path = unquote(parsed.path)[1:]
            # Parse the version ID the way the Java SDK does:
            # https://github.com/aws/aws-sdk-java/blob/master/aws-java-sdk-s3/src/main/java/com/amazonaws/services/s3/AmazonS3URI.java#L192
            query = parse_qs(parsed.query)
            version_id = query.pop("versionId", [None])[0]
            if query:
                raise URLParseError(f"Unexpected S3 query string: {parsed.query!r}")
            return cls(bucket, path, version_id)
        elif parsed.scheme == "file":
            if parsed.netloc not in ("", "localhost"):
                raise URLParseError("Unexpected hostname")
            if not parsed.path:
                raise URLParseError("Missing path")
            if not parsed.path.startswith("/"):
                raise URLParseError("Relative paths are not allowed")
            if parsed.query:
                raise URLParseError("Unexpected query")
            path = url2pathname(parsed.path)
            if parsed.path.endswith("/") and not path.endswith(os.path.sep):
                # On Windows, url2pathname loses the trailing `/`.
                path += os.path.sep
            return cls.from_path(path)
        else:
            raise URLParseError(f"Unexpected scheme: {parsed.scheme!r}")

    @classmethod
    def from_path(cls, path):
        path = os.fspath(path)
        new_path = os.path.realpath(path)
        # Use '/' as the path separator.
        if os.path.sep != "/":
            new_path = new_path.replace(os.path.sep, "/")
        # Add back a trailing '/' if the original path has it.
        if path.endswith(os.path.sep) or (
            os.path.altsep is not None and path.endswith(os.path.altsep)
        ):
            new_path += "/"
        return cls(None, new_path, None)

    def is_local(self):
        return self.bucket is None

    def looks_like_dir(self):
        return self.basename() == ""

    def join(self, rel_path):
        if self.version_id is not None:
            raise ValueError("Cannot append paths to URLs with a version ID")

        if os.name == "nt" and "\\" in rel_path:
            raise ValueError("Paths must use / as a separator")

        if self.path:
            new_path = self.path.rstrip("/") + "/" + rel_path.lstrip("/")
        else:
            new_path = rel_path.lstrip("/")
        return _PhysicalKey(self.bucket, new_path, None)

    def basename(self):
        return self.path.rsplit("/", 1)[-1]

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.bucket == other.bucket
            and self.path == other.path
            and self.version_id == other.version_id
        )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.bucket!r}, {self.path!r},"
            f" {self.version_id!r})"
        )

    def __str__(self):
        if self.bucket is None:
            return urlunparse(
                (
                    "file",
                    "",
                    pathname2url(self.path.replace("/", os.path.sep)),
                    None,
                    None,
                    None,
                )
            )
        else:
            if self.version_id is None:
                params = {}
            else:
                params = {"versionId": self.version_id}
            return urlunparse(
                (
                    "s3",
                    self.bucket,
                    quote(self.path),
                    None,
                    urlencode(params),
                    None,
                )
            )


class _WorkerContext:
    def __init__(self, progress, done, submit_task, s3_client):
        self.progress = progress
        self.done = done
        self.submit_task = submit_task
        self.s3_client = s3_client


class _FileTransfer:
    def __init__(
        self,
        src: Union[str, PathLike],
        dest: Union[str, PathLike],
        s3_client: BaseClient,
        disable_tqdm: bool = False,
        tqdm_desc: Optional[str] = None,
    ):
        self.src = _PhysicalKey.from_pathlike(src)
        self.dest = _PhysicalKey.from_pathlike(dest)
        self.s3_client = s3_client
        self.tqdm_desc = tqdm_desc
        self.disable_tqdm = disable_tqdm

        if self.src.looks_like_dir():
            raise ValueError(f"Expected file, got a directory: {self.src}")

        size, version_id = _get_size_and_version(self.src, self.s3_client)

        self.transfer_size = size

        if self.dest.looks_like_dir():
            self.dest = self.dest.join(self.src.basename())
        if self.src.version_id is None:
            self.src = _PhysicalKey(self.src.bucket, self.src.path, version_id)
        if self.dest.version_id:
            raise ValueError("Cannot set version_id on destination")

    def start(self) -> str:
        self._initialize()
        try:
            return self._transfer()
        finally:
            self._progress.close()

    def _initialize(self):
        self._lock = Lock()
        self._futures = deque()
        self._stopped = False
        self._progress = tqdm(
            desc=self.tqdm_desc,
            total=self.transfer_size,
            unit="B",
            unit_scale=True,
            disable=self.disable_tqdm,
            smoothing=0.1,
        )
        self._result = None
        self._executor = ThreadPoolExecutor(MAX_CONCURRENCY)

    def _progress_callback(self, bytes_transferred: int):
        if self._stopped:
            raise Exception("Interrupted")
        with self._lock:
            self._progress.update(bytes_transferred)

    def _submit_task(self, func, *args):
        future = self._executor.submit(func, *args)
        self._futures.append(future)

    def _worker(self, src: _PhysicalKey, dest: _PhysicalKey):
        def done_callback(key: _PhysicalKey):
            self._result = key

        ctx = _WorkerContext(
            progress=self._progress_callback,
            done=done_callback,
            submit_task=self._submit_task,
            s3_client=self.s3_client,
        )

        fro = "local" if src.is_local() else "remote"
        to = "local" if dest.is_local() else "remote"
        direction = (fro, to)

        if direction == ("local", "remote"):
            _upload_or_copy_file(
                ctx, self.transfer_size, src.path, dest.bucket, dest.path
            )

        elif direction == ("remote", "local"):
            _download_file(
                ctx, self.transfer_size, src.bucket, src.path, src.version_id, dest.path
            )

        elif direction == ("remote", "remote"):
            _copy_remote_file(
                ctx,
                self.transfer_size,
                src.bucket,
                src.path,
                src.version_id,
                dest.bucket,
                dest.path,
            )
        else:
            raise NotImplementedError(
                "Transfer for this combination of src"
                f" and dest is not implemented, src={src}, dest={dest}"
            )

    def _transfer(self) -> str:
        try:
            self._submit_task(self._worker, self.src, self.dest)

            while True:
                with self._lock:
                    if not self._futures:
                        break
                    future = self._futures.popleft()

                if future.cancelled():
                    continue

                try:
                    future.result()
                except (Exception, KeyboardInterrupt) as e:
                    for f in self._futures:
                        f.cancel()
                    self._futures.clear()
                    raise e
        finally:
            # Make sure all tasks exit quickly if the
            # main thread exits before they're done.
            self._stopped = True

        return str(self._result)


def _get_size_and_version(pk: _PhysicalKey, s3_client: BaseClient):
    """
    Gets size and version for the object at a given URL.

    Returns:
        size, version(str)
    """
    if pk.looks_like_dir():
        raise Exception("Invalid path: %r; cannot be a directory" % pk.path)

    version = None
    if pk.is_local():
        src_file = pathlib.Path(pk.path)
        if not src_file.is_file():
            raise Exception("Not a file: %r" % str(src_file))
        size = src_file.stat().st_size
    else:
        params = dict(Bucket=pk.bucket, Key=pk.path)
        if pk.version_id is not None:
            params.update(VersionId=pk.version_id)
        resp = s3_client.head_object(**params)
        size = resp["ContentLength"]
        version = resp.get("VersionId")
    return size, version


def _read_file_chunks(file, chunksize=s3_transfer_config.io_chunksize):
    return itertools.takewhile(bool, map(file.read, itertools.repeat(chunksize)))


def _calculate_etag(file_path: str) -> str:
    """
    Attempts to calculate a local file's ETag the way S3 does:
    - Normal uploads: MD5 of the file
    - Multi-part uploads: MD5 of the (binary) MD5s of the parts, dash, number of parts

    We can't know how the file was actually uploaded - but we're assuming it was done
    using the default settings, which we get from `s3_transfer_config`.
    """
    size = pathlib.Path(file_path).stat().st_size
    with open(file_path, "rb") as fd:
        if size < s3_transfer_config.multipart_threshold:
            contents = fd.read()
            etag = hashlib.md5(contents).hexdigest()
        else:
            adjuster = ChunksizeAdjuster()
            chunksize = adjuster.adjust_chunksize(
                s3_transfer_config.multipart_chunksize, size
            )

            hashes = []
            for contents in _read_file_chunks(fd, chunksize):
                hashes.append(hashlib.md5(contents).digest())
            etag = "%s-%d" % (hashlib.md5(b"".join(hashes)).hexdigest(), len(hashes))
    return '"%s"' % etag


def _upload_or_copy_file(
    ctx: _WorkerContext, size: int, src_path: str, dest_bucket: str, dest_path: str
):
    # Optimization: check if the remote file already exists and has the right ETag,
    # and skip the upload.
    if size >= UPLOAD_ETAG_OPTIMIZATION_THRESHOLD:
        try:
            params = dict(Bucket=dest_bucket, Key=dest_path)
            resp = ctx.s3_client.head_object(**params)
        except ClientError:
            # Destination doesn't exist, so fall through to the normal upload.
            pass
        else:
            # Check the ETag.
            dest_size = resp["ContentLength"]
            dest_etag = resp["ETag"]
            dest_version_id = resp.get("VersionId")
            if size == dest_size and resp.get("ServerSideEncryption") != "aws:kms":
                src_etag = _calculate_etag(src_path)
                if src_etag == dest_etag:
                    # Nothing more to do. We should not attempt to copy the object because
                    # that would cause the "copy object to itself" error.
                    ctx.progress(size)
                    ctx.done(_PhysicalKey(dest_bucket, dest_path, dest_version_id))
                    return  # Optimization succeeded.

    # If the optimization didn't happen, do the normal upload.
    _upload_file(ctx, size, src_path, dest_bucket, dest_path)


def _upload_file(
    ctx: _WorkerContext, size: int, src_path: str, dest_bucket: str, dest_key: str
):
    """
    For a small file, uploads it to s3 directly

    For a large file, creates multiple smaller tasks and
    submits them using `ctx.submit_task` to parent executor
    """
    if size < s3_transfer_config.multipart_threshold:
        with OSUtils().open_file_chunk_reader(src_path, 0, size, callbacks=[]) as fd:
            resp = ctx.s3_client.put_object(
                Body=fd,
                Bucket=dest_bucket,
                Key=dest_key,
            )

        version_id = resp.get("VersionId")  # Absent in unversioned buckets.
        ctx.progress(size)
        ctx.done(_PhysicalKey(dest_bucket, dest_key, version_id))
    else:
        resp = ctx.s3_client.create_multipart_upload(
            Bucket=dest_bucket,
            Key=dest_key,
        )
        upload_id = resp["UploadId"]

        adjuster = ChunksizeAdjuster()
        chunksize = adjuster.adjust_chunksize(
            s3_transfer_config.multipart_chunksize, size
        )

        chunk_offsets = list(range(0, size, chunksize))

        lock = Lock()
        remaining = len(chunk_offsets)
        parts = [None] * remaining

        def upload_part(i, start, end):
            nonlocal remaining
            part_id = i + 1
            with OSUtils().open_file_chunk_reader(
                src_path, start, end - start, [ctx.progress]
            ) as fd:
                part = ctx.s3_client.upload_part(
                    Body=fd,
                    Bucket=dest_bucket,
                    Key=dest_key,
                    UploadId=upload_id,
                    PartNumber=part_id,
                )
            with lock:
                ctx.progress(end - start)
                parts[i] = {"PartNumber": part_id, "ETag": part["ETag"]}
                remaining -= 1
                done = remaining == 0

            if done:
                resp = ctx.s3_client.complete_multipart_upload(
                    Bucket=dest_bucket,
                    Key=dest_key,
                    UploadId=upload_id,
                    MultipartUpload={"Parts": parts},
                )
                version_id = resp.get("VersionId")  # Absent in unversioned buckets.
                ctx.done(_PhysicalKey(dest_bucket, dest_key, version_id))

        for i, start in enumerate(chunk_offsets):
            end = min(start + chunksize, size)
            ctx.submit_task(upload_part, i, start, end)


def _copy_remote_file(
    ctx: _WorkerContext,
    size: int,
    src_bucket: str,
    src_key: str,
    src_version: str,
    dest_bucket: str,
    dest_key: str,
    extra_args=None,
):
    src_params = dict(Bucket=src_bucket, Key=src_key)
    if src_version is not None:
        src_params.update(VersionId=src_version)

    if size < s3_transfer_config.multipart_threshold:
        params = dict(
            CopySource=src_params,
            Bucket=dest_bucket,
            Key=dest_key,
        )

        if extra_args:
            params.update(extra_args)

        resp = ctx.s3_client.copy_object(**params)
        ctx.progress(size)
        version_id = resp.get("VersionId")  # Absent in unversioned buckets.
        ctx.done(_PhysicalKey(dest_bucket, dest_key, version_id))
    else:
        resp = ctx.s3_client.create_multipart_upload(
            Bucket=dest_bucket,
            Key=dest_key,
        )
        upload_id = resp["UploadId"]

        adjuster = ChunksizeAdjuster()
        chunksize = adjuster.adjust_chunksize(
            s3_transfer_config.multipart_chunksize, size
        )

        chunk_offsets = list(range(0, size, chunksize))

        lock = Lock()
        remaining = len(chunk_offsets)
        parts = [None] * remaining

        def upload_part(i, start, end):
            nonlocal remaining
            part_id = i + 1
            part = ctx.s3_client.upload_part_copy(
                CopySource=src_params,
                CopySourceRange=f"bytes={start}-{end-1}",
                Bucket=dest_bucket,
                Key=dest_key,
                UploadId=upload_id,
                PartNumber=part_id,
            )
            with lock:
                parts[i] = {
                    "PartNumber": part_id,
                    "ETag": part["CopyPartResult"]["ETag"],
                }
                remaining -= 1
                done = remaining == 0

            ctx.progress(end - start)

            if done:
                resp = ctx.s3_client.complete_multipart_upload(
                    Bucket=dest_bucket,
                    Key=dest_key,
                    UploadId=upload_id,
                    MultipartUpload={"Parts": parts},
                )
                version_id = resp.get("VersionId")  # Absent in unversioned buckets.
                ctx.done(_PhysicalKey(dest_bucket, dest_key, version_id))

        for i, start in enumerate(chunk_offsets):
            end = min(start + chunksize, size)
            ctx.submit_task(upload_part, i, start, end)


def _download_file(
    ctx: _WorkerContext,
    size: int,
    src_bucket: str,
    src_key: str,
    src_version: str,
    dest_path: str,
):
    """
    Downloads S3 file to local
    """
    dest_file = pathlib.Path(dest_path)
    if dest_file.is_reserved():
        raise ValueError("Cannot download to %r: reserved file name" % dest_path)

    params = dict(Bucket=src_bucket, Key=src_key)

    dest_file.parent.mkdir(parents=True, exist_ok=True)

    with dest_file.open("wb") as f:
        fileno = f.fileno()
        is_regular_file = stat.S_ISREG(os.stat(fileno).st_mode)

    if src_version is not None:
        params.update(VersionId=src_version)

    part_size = s3_transfer_config.multipart_chunksize
    is_multi_part = (
        is_regular_file
        and size >= s3_transfer_config.multipart_threshold
        and size > part_size
    )
    part_numbers = range(math.ceil(size / part_size)) if is_multi_part else (None,)
    remaining_counter = len(part_numbers)
    remaining_counter_lock = Lock()

    def download_part(part_number):
        nonlocal remaining_counter

        with dest_file.open("r+b") as chunk_f:
            if part_number is not None:
                start = part_number * part_size
                end = min(start + part_size, size) - 1
                part_params = dict(params, Range=f"bytes={start}-{end}")
                chunk_f.seek(start)
            else:
                part_params = params

            resp = ctx.s3_client.get_object(**part_params)
            body = resp["Body"]
            while True:
                chunk = body.read(s3_transfer_config.io_chunksize)
                if not chunk:
                    break
                ctx.progress(chunk_f.write(chunk))

        with remaining_counter_lock:
            remaining_counter -= 1
            done = remaining_counter == 0
        if done:
            ctx.done(_PhysicalKey.from_path(dest_path))

    for part_number in part_numbers:
        ctx.submit_task(download_part, part_number)



import copy
from collections import deque
from decimal import Decimal
import json
import os
import textwrap

import pytest

import glom


def _get_codeblock(lines, offset):
    if lines[offset:offset + 2] != [".. code-block:: python\n", "\n"]:
        return None
    start = offset + 2
    try:
        finish = lines.index('\n', start)
    except ValueError:
        return None
    return textwrap.dedent("".join(lines[start:finish]))


def _find_snippets():
    path = os.path.dirname(os.path.abspath(__file__)) + '/../../docs/snippets.rst'
    with open(path, 'r') as snippet_file:
        lines = list(snippet_file)
    snippets = []
    for line_no in range(len(lines)):
        source = _get_codeblock(lines, line_no)
        if source:
            snippets.append((line_no, source))
    return snippets


try:
    SNIPPETS = _find_snippets()
except:
    SNIPPETS = []  # in case running in an environment without docs

SNIPPETS_GLOBALS = copy.copy(glom.__dict__)
SNIPPETS_GLOBALS.update(dict(
    json=json,
    deque=deque,
    Decimal=Decimal,
    data=json.dumps({'a': ['b']}),
    contacts=[{'primary_email': {'email': 'a@example.com'}}, {}],
    glom=glom.glom))


@pytest.mark.parametrize("line,source", SNIPPETS)
def test_snippet(line, source):
    if '>>>' in source:
        return  # maybe doctest output checking
    code = compile(source, 'snippets.rst', 'exec')
    if 'django' in source:
        return  # maybe in the future
    eval(code, SNIPPETS_GLOBALS)




# Based on the example from hello.py
import altair as alt
from vega_datasets import data

import datapane as dp
import datapane.blocks.asset
import datapane.blocks.layout
import datapane.blocks.text

source = data.cars()

plot1 = (
    alt.Chart(source)
    .mark_circle(size=60)
    .encode(
        x="Horsepower",
        y="Miles_per_Gallon",
        color="Origin",
        tooltip=["Name", "Origin", "Horsepower", "Miles_per_Gallon"],
    )
    .interactive()
)

report = dp.App(
    """🎆 Congratulations on uploading your first Datapane report!! 🎆""",
    datapane.blocks.inline_text.Formula("x^2 + y^2 = z^2"),
    datapane.blocks.layout.Group(
        dp.BigNumber(heading="Number of percentage points", value="84%", change="2%", is_upward_change=True),
        dp.BigNumber(heading="Simple Statistic", value=100),
        columns=2,
    ),
    datapane.blocks.layout.Select(
        datapane.blocks.asset.Plot(plot1, label="Chart"), datapane.blocks.asset.DataTable(source, label="Data")
    ),
)

report.upload(name="Hello World", description="My first Datapane Report", tags=["example"], open=True)



from __future__ import division

from pytest import raises

from glom import glom, T, SKIP, STOP, Auto, BadSpec, Val
from glom.grouping import Group, First, Avg, Max, Min, Sample, Limit

from glom.reduction import Merge, Flatten, Sum, Count


def test_bucketing():
    assert glom(range(4), Group({T % 2: [T]})) == {0: [0, 2], 1: [1, 3]}
    assert (glom(range(6), Group({T % 3: {T % 2: [T / 10.0]}})) ==
        {0: {0: [0.0], 1: [0.3]}, 1: {1: [0.1], 0: [0.4]}, 2: {0: [0.2], 1: [0.5]}})



def test_corner_cases():
    target = range(5)

    # immediate stop dict
    assert glom(target, Group({Val(STOP): [T]})) == {}

    # immediate stop list
    assert glom(target, Group([Val(STOP)])) == []

    # dict key SKIP
    assert glom(target, Group({(lambda t: SKIP if t < 3 else t): T})) == {3: 3, 4: 4}

    # dict val SKIP
    assert glom(target, Group({T: lambda t: t if t % 2 else SKIP})) == {3: 3, 1: 1}

    # list val SKIP
    assert glom(target, Group([lambda t: t if t % 2 else SKIP])) == [1, 3]

    # embedded auto spec (lol @ 0 being 0 bit length)
    assert glom(target, Group({Auto(('bit_length', T())): [T]})) == {0: [0], 1: [1], 2: [2, 3], 3: [4]}

    # no dicts inside lists in Group mode
    with raises(BadSpec):
        assert glom(target, Group([{T: T}]))

    # check only supported types
    with raises(BadSpec):
        assert glom(target, Group('no string support yet'))

    # bucket ints by their bit length and then odd/even, limited to 3 per bucket
    spec = Group({T.bit_length(): {lambda t: t % 2: Limit(3)}})
    res = glom(range(20), spec)
    assert res == {0: {0: [0]},
                   1: {1: [1]},
                   2: {0: [2], 1: [3]},
                   3: {0: [4, 6], 1: [5, 7]},
                   4: {0: [8, 10, 12], 1: [9, 11, 13]}}

    return


def test_agg():
    t = list(range(10))
    assert glom(t, Group(First())) == 0
    assert glom(t, Group(T)) == 9  # this is basically Last

    assert glom(t, Group(Avg())) == sum(t) / len(t)
    assert glom(t, Group(Sum())) == sum(t)

    assert glom([0, 1, 0], Group(Max())) == 1
    assert glom([1, 0, 1], Group(Min())) == 0

    assert repr(Group(First())) == 'Group(First())'
    assert repr(Avg()) == 'Avg()'
    assert repr(Max()) == 'Max()'
    assert repr(Min()) == 'Min()'
    assert repr(Sum()) == 'Sum()'
    assert repr(Count()) == 'Count()'

    assert glom(range(10), Group({lambda t: t % 2: Count()})) == {
		0: 5, 1: 5}


def test_limit():
    t = list(range(10))
    assert glom(t, Group(Limit(1, T))) == 0
    assert glom(t, Group(Limit(3, Max()))) == 2
    assert glom(t, Group(Limit(3, [T]))) == [0, 1, 2]

    assert repr(Group(Limit(3, Max()))) == 'Group(Limit(3, Max()))'

    with raises(BadSpec):
        assert glom(t, Limit(1))  # needs to be wrapped in Group for now
    return


def test_reduce():
    assert glom([[1], [2, 3]], Group(Flatten())) == [1, 2, 3]
    assert glom([{'a': 1}, {'b': 2}], Group(Merge())) == {'a': 1, 'b': 2}
    assert glom([[[1]], [[2, 3], [4]]], Group(Flatten(Flatten()))) == [1, 2, 3, 4]


def test_sample():
    spec = Group(Sample(5))
    assert glom([1, 2, 3], spec) == [1, 2, 3]

    assert repr(spec) == 'Group(Sample(5))'

    s = glom([1, 2, 3], Group(Sample(2)))
    assert s in [[1, 2], [1, 3], [2, 1], [2, 3], [3, 1], [3, 2]]



import base64
import json
import logging
import re
import traceback
import warnings
import zlib
from abc import ABC, abstractmethod
from enum import Enum
from functools import partial
from http import HTTPStatus
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Match,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from aws_lambda_powertools.event_handler import content_types
from aws_lambda_powertools.event_handler.exceptions import NotFoundError, ServiceError
from aws_lambda_powertools.event_handler.openapi.constants import DEFAULT_API_VERSION, DEFAULT_OPENAPI_VERSION
from aws_lambda_powertools.event_handler.openapi.exceptions import RequestValidationError
from aws_lambda_powertools.event_handler.openapi.swagger_ui.html import generate_swagger_html
from aws_lambda_powertools.event_handler.openapi.types import (
    COMPONENT_REF_PREFIX,
    METHODS_WITH_BODY,
    validation_error_definition,
    validation_error_response_definition,
)
from aws_lambda_powertools.event_handler.util import _FrozenDict
from aws_lambda_powertools.shared.cookies import Cookie
from aws_lambda_powertools.shared.functions import powertools_dev_is_set
from aws_lambda_powertools.shared.json_encoder import Encoder
from aws_lambda_powertools.shared.types import Literal
from aws_lambda_powertools.utilities.data_classes import (
    ALBEvent,
    APIGatewayProxyEvent,
    APIGatewayProxyEventV2,
    BedrockAgentEvent,
    LambdaFunctionUrlEvent,
    VPCLatticeEvent,
    VPCLatticeEventV2,
)
from aws_lambda_powertools.utilities.data_classes.common import BaseProxyEvent
from aws_lambda_powertools.utilities.typing import LambdaContext

logger = logging.getLogger(__name__)

_DYNAMIC_ROUTE_PATTERN = r"(<\w+>)"
_SAFE_URI = "-._~()'!*:@,;=+&$"  # https://www.ietf.org/rfc/rfc3986.txt
# API GW/ALB decode non-safe URI chars; we must support them too
_UNSAFE_URI = r"%<> \[\]{}|^"
_NAMED_GROUP_BOUNDARY_PATTERN = rf"(?P\1[{_SAFE_URI}{_UNSAFE_URI}\\w]+)"
_DEFAULT_OPENAPI_RESPONSE_DESCRIPTION = "Successful Response"
_ROUTE_REGEX = "^{}$"

ResponseEventT = TypeVar("ResponseEventT", bound=BaseProxyEvent)
ResponseT = TypeVar("ResponseT")

if TYPE_CHECKING:
    from aws_lambda_powertools.event_handler.openapi.compat import (
        JsonSchemaValue,
        ModelField,
    )
    from aws_lambda_powertools.event_handler.openapi.models import (
        Contact,
        License,
        OpenAPI,
        Server,
    )
    from aws_lambda_powertools.event_handler.openapi.params import Dependant
    from aws_lambda_powertools.event_handler.openapi.types import (
        TypeModelOrEnum,
    )


class ProxyEventType(Enum):
    """An enumerations of the supported proxy event types."""

    APIGatewayProxyEvent = "APIGatewayProxyEvent"
    APIGatewayProxyEventV2 = "APIGatewayProxyEventV2"
    ALBEvent = "ALBEvent"
    BedrockAgentEvent = "BedrockAgentEvent"
    VPCLatticeEvent = "VPCLatticeEvent"
    VPCLatticeEventV2 = "VPCLatticeEventV2"
    LambdaFunctionUrlEvent = "LambdaFunctionUrlEvent"


class CORSConfig:
    """CORS Config

    Examples
    --------

    Simple cors example using the default permissive cors, not this should only be used during early prototyping

    ```python
    from aws_lambda_powertools.event_handler import APIGatewayRestResolver

    app = APIGatewayRestResolver()

    @app.get("/my/path", cors=True)
    def with_cors():
        return {"message": "Foo"}
    ```

    Using a custom CORSConfig where `with_cors` used the custom provided CORSConfig and `without_cors`
    do not include any cors headers.

    ```python
    from aws_lambda_powertools.event_handler.api_gateway import (
        APIGatewayRestResolver, CORSConfig
    )

    cors_config = CORSConfig(
        allow_origin="https://wwww.example.com/",
        extra_origins=["https://dev.example.com/"],
        expose_headers=["x-exposed-response-header"],
        allow_headers=["x-custom-request-header"],
        max_age=100,
        allow_credentials=True,
    )
    app = APIGatewayRestResolver(cors=cors_config)

    @app.get("/my/path")
    def with_cors():
        return {"message": "Foo"}

    @app.get("/another-one", cors=False)
    def without_cors():
        return {"message": "Foo"}
    ```
    """

    _REQUIRED_HEADERS = ["Authorization", "Content-Type", "X-Amz-Date", "X-Api-Key", "X-Amz-Security-Token"]

    def __init__(
        self,
        allow_origin: str = "*",
        extra_origins: Optional[List[str]] = None,
        allow_headers: Optional[List[str]] = None,
        expose_headers: Optional[List[str]] = None,
        max_age: Optional[int] = None,
        allow_credentials: bool = False,
    ):
        """
        Parameters
        ----------
        allow_origin: str
            The value of the `Access-Control-Allow-Origin` to send in the response. Defaults to "*", but should
            only be used during development.
        extra_origins: Optional[List[str]]
            The list of additional allowed origins.
        allow_headers: Optional[List[str]]
            The list of additional allowed headers. This list is added to list of
            built-in allowed headers: `Authorization`, `Content-Type`, `X-Amz-Date`,
            `X-Api-Key`, `X-Amz-Security-Token`.
        expose_headers: Optional[List[str]]
            A list of values to return for the Access-Control-Expose-Headers
        max_age: Optional[int]
            The value for the `Access-Control-Max-Age`
        allow_credentials: bool
            A boolean value that sets the value of `Access-Control-Allow-Credentials`
        """
        self._allowed_origins = [allow_origin]
        if extra_origins:
            self._allowed_origins.extend(extra_origins)
        self.allow_headers = set(self._REQUIRED_HEADERS + (allow_headers or []))
        self.expose_headers = expose_headers or []
        self.max_age = max_age
        self.allow_credentials = allow_credentials

    def to_dict(self, origin: Optional[str]) -> Dict[str, str]:
        """Builds the configured Access-Control http headers"""

        # If there's no Origin, don't add any CORS headers
        if not origin:
            return {}

        # If the origin doesn't match any of the allowed origins, and we don't allow all origins ("*"),
        # don't add any CORS headers
        if origin not in self._allowed_origins and "*" not in self._allowed_origins:
            return {}

        # The origin matched an allowed origin, so return the CORS headers
        headers: Dict[str, str] = {
            "Access-Control-Allow-Origin": origin,
            "Access-Control-Allow-Headers": ",".join(sorted(self.allow_headers)),
        }

        if self.expose_headers:
            headers["Access-Control-Expose-Headers"] = ",".join(self.expose_headers)
        if self.max_age is not None:
            headers["Access-Control-Max-Age"] = str(self.max_age)
        if self.allow_credentials is True:
            headers["Access-Control-Allow-Credentials"] = "true"
        return headers


class Response(Generic[ResponseT]):
    """Response data class that provides greater control over what is returned from the proxy event"""

    def __init__(
        self,
        status_code: int,
        content_type: Optional[str] = None,
        body: Optional[ResponseT] = None,
        headers: Optional[Dict[str, Union[str, List[str]]]] = None,
        cookies: Optional[List[Cookie]] = None,
        compress: Optional[bool] = None,
    ):
        """

        Parameters
        ----------
        status_code: int
            Http status code, example 200
        content_type: str
            Optionally set the Content-Type header, example "application/json". Note this will be merged into any
            provided http headers
        body: Union[str, bytes, None]
            Optionally set the response body. Note: bytes body will be automatically base64 encoded
        headers: dict[str, Union[str, List[str]]]
            Optionally set specific http headers. Setting "Content-Type" here would override the `content_type` value.
        cookies: list[Cookie]
            Optionally set cookies.
        """
        self.status_code = status_code
        self.body = body
        self.base64_encoded = False
        self.headers: Dict[str, Union[str, List[str]]] = headers if headers else {}
        self.cookies = cookies or []
        self.compress = compress
        self.content_type = content_type
        if content_type:
            self.headers.setdefault("Content-Type", content_type)

    def is_json(self) -> bool:
        """
        Returns True if the response is JSON, based on the Content-Type.
        """
        content_type = self.headers.get("Content-Type", "")
        if isinstance(content_type, list):
            content_type = content_type[0]
        return content_type.startswith("application/json")


class Route:
    """Internally used Route Configuration"""

    def __init__(
        self,
        method: str,
        path: str,
        rule: Pattern,
        func: Callable,
        cors: bool,
        compress: bool,
        cache_control: Optional[str],
        summary: Optional[str],
        description: Optional[str],
        responses: Optional[Dict[int, Dict[str, Any]]],
        response_description: Optional[str],
        tags: Optional[List[str]],
        operation_id: Optional[str],
        include_in_schema: bool,
        middlewares: Optional[List[Callable[..., Response]]],
    ):
        """

        Parameters
        ----------

        method: str
            The HTTP method, example "GET"
        path: str
            The path of the route
        rule: Pattern
            The route rule, example "/my/path"
        func: Callable
            The route handler function
        cors: bool
            Whether or not to enable CORS for this route
        compress: bool
            Whether or not to enable gzip compression for this route
        cache_control: Optional[str]
            The cache control header value, example "max-age=3600"
        summary: Optional[str]
            The OpenAPI summary for this route
        description: Optional[str]
            The OpenAPI description for this route
        responses: Optional[Dict[int, Dict[str, Any]]]
            The OpenAPI responses for this route
        response_description: Optional[str]
            The OpenAPI response description for this route
        tags: Optional[List[str]]
            The list of OpenAPI tags to be used for this route
        operation_id: Optional[str]
            The OpenAPI operationId for this route
        include_in_schema: bool
            Whether or not to include this route in the OpenAPI schema
        middlewares: Optional[List[Callable[..., Response]]]
            The list of route middlewares to be called in order.
        """
        self.method = method.upper()
        self.path = "/" if path.strip() == "" else path

        # OpenAPI spec only understands paths with { }. So we'll have to convert Powertools' < >.
        # https://swagger.io/specification/#path-templating
        self.openapi_path = re.sub(r"<(.*?)>", lambda m: f"{{{''.join(m.group(1))}}}", self.path)

        self.rule = rule
        self.func = func
        self._middleware_stack = func
        self.cors = cors
        self.compress = compress
        self.cache_control = cache_control
        self.summary = summary
        self.description = description
        self.responses = responses
        self.response_description = response_description
        self.tags = tags or []
        self.include_in_schema = include_in_schema
        self.middlewares = middlewares or []
        self.operation_id = operation_id or self._generate_operation_id()

        # _middleware_stack_built is used to ensure the middleware stack is only built once.
        self._middleware_stack_built = False

        # _dependant is used to cache the dependant model for the handler function
        self._dependant: Optional["Dependant"] = None

        # _body_field is used to cache the dependant model for the body field
        self._body_field: Optional["ModelField"] = None

    def __call__(
        self,
        router_middlewares: List[Callable],
        app: "ApiGatewayResolver",
        route_arguments: Dict[str, str],
    ) -> Union[Dict, Tuple, Response]:
        """Calling the Router class instance will trigger the following actions:
            1. If Route Middleware stack has not been built, build it
            2. Call the Route Middleware stack wrapping the original function
                handler with the app and route arguments.

        Parameters
        ----------
        router_middlewares: List[Callable]
            The list of Router Middlewares (assigned to ALL routes)
        app: "ApiGatewayResolver"
            The ApiGatewayResolver instance to pass into the middleware stack
        route_arguments: Dict[str, str]
            The route arguments to pass to the app function (extracted from the Api Gateway
            Lambda Message structure from AWS)

        Returns
        -------
        Union[Dict, Tuple, Response]
            API Response object in ALL cases, except when the original API route
            handler is called which may also return a Dict, Tuple, or Response.
        """

        # Save CPU cycles by building middleware stack once
        if not self._middleware_stack_built:
            self._build_middleware_stack(router_middlewares=router_middlewares)

        # If debug is turned on then output the middleware stack to the console
        if app._debug:
            print(f"\nProcessing Route:::{self.func.__name__} ({app.context['_path']})")
            # Collect ALL middleware for debug printing - include internal _registered_api_adapter
            all_middlewares = router_middlewares + self.middlewares + [_registered_api_adapter]
            print("\nMiddleware Stack:")
            print("=================")
            print("\n".join(getattr(item, "__name__", "Unknown") for item in all_middlewares))
            print("=================")

        # Add Route Arguments to app context
        app.append_context(_route_args=route_arguments)

        # Call the Middleware Wrapped _call_stack function handler with the app
        return self._middleware_stack(app)

    def _build_middleware_stack(self, router_middlewares: List[Callable[..., Any]]) -> None:
        """
        Builds the middleware stack for the handler by wrapping each
        handler in an instance of MiddlewareWrapper which is used to contain the state
        of each middleware step.

        Middleware is represented by a standard Python Callable construct.  Any Middleware
        handler wanting to short-circuit the middlware call chain can raise an exception
        to force the Python call stack created by the handler call-chain to naturally un-wind.

        This becomes a simple concept for developers to understand and reason with - no additional
        gymanstics other than plain old try ... except.

        Notes
        -----
        The Route Middleware stack is processed in reverse order. This is so the stack of
        middleware handlers is applied in the order of being added to the handler.
        """
        all_middlewares = router_middlewares + self.middlewares
        logger.debug(f"Building middleware stack: {all_middlewares}")

        # IMPORTANT:
        # this must be the last middleware in the stack (tech debt for backward
        # compatibility purposes)
        #
        # This adapter will:
        #   1. Call the registered API passing only the expected route arguments extracted from the path
        # and not the middleware.
        #   2. Adapt the response type of the route handler (Union[Dict, Tuple, Response])
        # and normalise into a Response object so middleware will always have a constant signature
        all_middlewares.append(_registered_api_adapter)

        # Wrap the original route handler function in the middleware handlers
        # using the MiddlewareWrapper class callable construct in reverse order to
        # ensure middleware is applied in the order the user defined.
        #
        # Start with the route function and wrap from last to the first Middleware handler.
        for handler in reversed(all_middlewares):
            self._middleware_stack = MiddlewareFrame(current_middleware=handler, next_middleware=self._middleware_stack)

        self._middleware_stack_built = True

    @property
    def dependant(self) -> "Dependant":
        if self._dependant is None:
            from aws_lambda_powertools.event_handler.openapi.dependant import get_dependant

            self._dependant = get_dependant(path=self.openapi_path, call=self.func)

        return self._dependant

    @property
    def body_field(self) -> Optional["ModelField"]:
        if self._body_field is None:
            from aws_lambda_powertools.event_handler.openapi.dependant import get_body_field

            self._body_field = get_body_field(dependant=self.dependant, name=self.operation_id)

        return self._body_field

    def _get_openapi_path(
        self,
        *,
        dependant: "Dependant",
        operation_ids: Set[str],
        model_name_map: Dict["TypeModelOrEnum", str],
        field_mapping: Dict[Tuple["ModelField", Literal["validation", "serialization"]], "JsonSchemaValue"],
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Returns the OpenAPI path and definitions for the route.
        """
        from aws_lambda_powertools.event_handler.openapi.dependant import get_flat_params

        path = {}
        definitions: Dict[str, Any] = {}

        # Gather all the route parameters
        operation = self._openapi_operation_metadata(operation_ids=operation_ids)
        parameters: List[Dict[str, Any]] = []
        all_route_params = get_flat_params(dependant)
        operation_params = self._openapi_operation_parameters(
            all_route_params=all_route_params,
            model_name_map=model_name_map,
            field_mapping=field_mapping,
        )
        parameters.extend(operation_params)

        # Add the parameters to the OpenAPI operation
        if parameters:
            all_parameters = {(param["in"], param["name"]): param for param in parameters}
            required_parameters = {(param["in"], param["name"]): param for param in parameters if param.get("required")}
            all_parameters.update(required_parameters)
            operation["parameters"] = list(all_parameters.values())

        # Add the request body to the OpenAPI operation, if applicable
        if self.method.upper() in METHODS_WITH_BODY:
            request_body_oai = self._openapi_operation_request_body(
                body_field=self.body_field,
                model_name_map=model_name_map,
                field_mapping=field_mapping,
            )
            if request_body_oai:
                operation["requestBody"] = request_body_oai

        # Add the response to the OpenAPI operation
        if self.responses:
            # If the user supplied responses, we use them and don't set a default 200 response
            operation["responses"] = self.responses
        else:
            # Set the default 200 response
            responses = operation.setdefault("responses", self.responses or {})
            success_response = responses.setdefault(200, {})
            success_response["description"] = self.response_description or _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION
            success_response["content"] = {"application/json": {"schema": {}}}
            json_response = success_response["content"].setdefault("application/json", {})

            # Add the response schema to the OpenAPI 200 response
            json_response.update(
                self._openapi_operation_return(
                    param=dependant.return_param,
                    model_name_map=model_name_map,
                    field_mapping=field_mapping,
                ),
            )

            # Add validation failure response (422)
            operation["responses"][422] = {
                "description": "Validation Error",
                "content": {
                    "application/json": {
                        "schema": {"$ref": COMPONENT_REF_PREFIX + "HTTPValidationError"},
                    },
                },
            }

            # Add the validation error schema to the definitions, but only if it hasn't been added yet
            if "ValidationError" not in definitions:
                definitions.update(
                    {
                        "ValidationError": validation_error_definition,
                        "HTTPValidationError": validation_error_response_definition,
                    },
                )

        path[self.method.lower()] = operation

        # Generate the response schema
        return path, definitions

    def _openapi_operation_summary(self) -> str:
        """
        Returns the OpenAPI operation summary. If the user has not provided a summary, we
        generate one based on the route path and method.
        """
        return self.summary or f"{self.method.upper()} {self.openapi_path}"

    def _openapi_operation_metadata(self, operation_ids: Set[str]) -> Dict[str, Any]:
        """
        Returns the OpenAPI operation metadata. If the user has not provided a description, we
        generate one based on the route path and method.
        """
        operation: Dict[str, Any] = {}

        # Ensure tags is added to the operation
        if self.tags:
            operation["tags"] = self.tags

        # Ensure summary is added to the operation
        operation["summary"] = self._openapi_operation_summary()

        # Ensure description is added to the operation
        if self.description:
            operation["description"] = self.description

        # Ensure operationId is unique
        if self.operation_id in operation_ids:
            message = f"Duplicate Operation ID {self.operation_id} for function {self.func.__name__}"
            file_name = getattr(self.func, "__globals__", {}).get("__file__")
            if file_name:
                message += f" in {file_name}"
            warnings.warn(message, stacklevel=1)

        # Adds the operation
        operation_ids.add(self.operation_id)
        operation["operationId"] = self.operation_id

        return operation

    @staticmethod
    def _openapi_operation_request_body(
        *,
        body_field: Optional["ModelField"],
        model_name_map: Dict["TypeModelOrEnum", str],
        field_mapping: Dict[Tuple["ModelField", Literal["validation", "serialization"]], "JsonSchemaValue"],
    ) -> Optional[Dict[str, Any]]:
        """
        Returns the OpenAPI operation request body.
        """
        from aws_lambda_powertools.event_handler.openapi.compat import ModelField, get_schema_from_model_field
        from aws_lambda_powertools.event_handler.openapi.params import Body

        # Check that there is a body field and it's a Pydantic's model field
        if not body_field:
            return None

        if not isinstance(body_field, ModelField):
            raise AssertionError(f"Expected ModelField, got {body_field}")

        # Generate the request body schema
        body_schema = get_schema_from_model_field(
            field=body_field,
            model_name_map=model_name_map,
            field_mapping=field_mapping,
        )

        field_info = cast(Body, body_field.field_info)
        request_media_type = field_info.media_type
        required = body_field.required
        request_body_oai: Dict[str, Any] = {}
        if required:
            request_body_oai["required"] = required

        # Generate the request body media type
        request_media_content: Dict[str, Any] = {"schema": body_schema}
        request_body_oai["content"] = {request_media_type: request_media_content}
        return request_body_oai

    @staticmethod
    def _openapi_operation_parameters(
        *,
        all_route_params: Sequence["ModelField"],
        model_name_map: Dict["TypeModelOrEnum", str],
        field_mapping: Dict[
            Tuple["ModelField", Literal["validation", "serialization"]],
            "JsonSchemaValue",
        ],
    ) -> List[Dict[str, Any]]:
        """
        Returns the OpenAPI operation parameters.
        """
        from aws_lambda_powertools.event_handler.openapi.compat import (
            get_schema_from_model_field,
        )
        from aws_lambda_powertools.event_handler.openapi.params import Param

        parameters = []
        for param in all_route_params:
            field_info = param.field_info
            field_info = cast(Param, field_info)
            if not field_info.include_in_schema:
                continue

            param_schema = get_schema_from_model_field(
                field=param,
                model_name_map=model_name_map,
                field_mapping=field_mapping,
            )

            parameter = {
                "name": param.alias,
                "in": field_info.in_.value,
                "required": param.required,
                "schema": param_schema,
            }

            if field_info.description:
                parameter["description"] = field_info.description

            if field_info.deprecated:
                parameter["deprecated"] = field_info.deprecated

            parameters.append(parameter)

        return parameters

    @staticmethod
    def _openapi_operation_return(
        *,
        param: Optional["ModelField"],
        model_name_map: Dict["TypeModelOrEnum", str],
        field_mapping: Dict[
            Tuple["ModelField", Literal["validation", "serialization"]],
            "JsonSchemaValue",
        ],
    ) -> Dict[str, Any]:
        """
        Returns the OpenAPI operation return.
        """
        if param is None:
            return {}

        from aws_lambda_powertools.event_handler.openapi.compat import (
            get_schema_from_model_field,
        )

        return_schema = get_schema_from_model_field(
            field=param,
            model_name_map=model_name_map,
            field_mapping=field_mapping,
        )

        return {"schema": return_schema}

    def _generate_operation_id(self) -> str:
        operation_id = self.func.__name__ + self.openapi_path
        operation_id = re.sub(r"\W", "_", operation_id)
        operation_id = operation_id + "_" + self.method.lower()
        return operation_id


class ResponseBuilder(Generic[ResponseEventT]):
    """Internally used Response builder"""

    def __init__(
        self,
        response: Response,
        serializer: Callable[[Any], str] = partial(json.dumps, separators=(",", ":"), cls=Encoder),
        route: Optional[Route] = None,
    ):
        self.response = response
        self.serializer = serializer
        self.route = route

    def _add_cors(self, event: ResponseEventT, cors: CORSConfig):
        """Update headers to include the configured Access-Control headers"""
        self.response.headers.update(cors.to_dict(event.get_header_value("Origin")))

    def _add_cache_control(self, cache_control: str):
        """Set the specified cache control headers for 200 http responses. For non-200 `no-cache` is used."""
        cache_control = cache_control if self.response.status_code == 200 else "no-cache"
        self.response.headers["Cache-Control"] = cache_control

    @staticmethod
    def _has_compression_enabled(
        route_compression: bool,
        response_compression: Optional[bool],
        event: ResponseEventT,
    ) -> bool:
        """
        Checks if compression is enabled.

        NOTE: Response compression takes precedence.

        Parameters
        ----------
        route_compression: bool, optional
            A boolean indicating whether compression is enabled or not in the route setting.
        response_compression: bool, optional
            A boolean indicating whether compression is enabled or not in the response setting.
        event: ResponseEventT
            The event object containing the request details.

        Returns
        -------
        bool
            True if compression is enabled and the "gzip" encoding is accepted, False otherwise.
        """
        encoding: str = event.get_header_value(
            name="accept-encoding",
            default_value="",
            case_sensitive=False,
        )  # noqa: E501
        if "gzip" in encoding:
            if response_compression is not None:
                return response_compression  # e.g., Response(compress=False/True))
            if route_compression:
                return True  # e.g., @app.get(compress=True)

        return False

    def _compress(self):
        """Compress the response body, but only if `Accept-Encoding` headers includes gzip."""
        self.response.headers["Content-Encoding"] = "gzip"
        if isinstance(self.response.body, str):
            logger.debug("Converting string response to bytes before compressing it")
            self.response.body = bytes(self.response.body, "utf-8")
        gzip = zlib.compressobj(9, zlib.DEFLATED, zlib.MAX_WBITS | 16)
        self.response.body = gzip.compress(self.response.body) + gzip.flush()

    def _route(self, event: ResponseEventT, cors: Optional[CORSConfig]):
        """Optionally handle any of the route's configure response handling"""
        if self.route is None:
            return
        if self.route.cors:
            self._add_cors(event, cors or CORSConfig())
        if self.route.cache_control:
            self._add_cache_control(self.route.cache_control)
        if self._has_compression_enabled(
            route_compression=self.route.compress,
            response_compression=self.response.compress,
            event=event,
        ):
            self._compress()

    def build(self, event: ResponseEventT, cors: Optional[CORSConfig] = None) -> Dict[str, Any]:
        """Build the full response dict to be returned by the lambda"""

        # We only apply the serializer when the content type is JSON and the
        # body is not a str, to avoid double encoding
        if self.response.is_json() and not isinstance(self.response.body, str):
            self.response.body = self.serializer(self.response.body)

        self._route(event, cors)

        if isinstance(self.response.body, bytes):
            logger.debug("Encoding bytes response with base64")
            self.response.base64_encoded = True
            self.response.body = base64.b64encode(self.response.body).decode()

        return {
            "statusCode": self.response.status_code,
            "body": self.response.body,
            "isBase64Encoded": self.response.base64_encoded,
            **event.header_serializer().serialize(headers=self.response.headers, cookies=self.response.cookies),
        }


class BaseRouter(ABC):
    current_event: BaseProxyEvent
    lambda_context: LambdaContext
    context: dict
    _router_middlewares: List[Callable] = []
    processed_stack_frames: List[str] = []

    @abstractmethod
    def route(
        self,
        rule: str,
        method: Any,
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: str = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        raise NotImplementedError()

    def use(self, middlewares: List[Callable[..., Response]]) -> None:
        """
        Add one or more global middlewares that run before/after route specific middleware.

        NOTE: Middlewares are called in insertion order.

        Parameters
        ----------
        middlewares: List[Callable[..., Response]]
            List of global middlewares to be used

        Examples
        --------

        Add middlewares to be used for every request processed by the Router.

        ```python
        from aws_lambda_powertools import Logger
        from aws_lambda_powertools.event_handler import APIGatewayRestResolver, Response
        from aws_lambda_powertools.event_handler.middlewares import NextMiddleware

        logger = Logger()
        app = APIGatewayRestResolver()

        def log_request_response(app: APIGatewayRestResolver, next_middleware: NextMiddleware) -> Response:
            logger.info("Incoming request", path=app.current_event.path, request=app.current_event.raw_event)

            result = next_middleware(app)
            logger.info("Response received", response=result.__dict__)

            return result

        app.use(middlewares=[log_request_response])


        def lambda_handler(event, context):
            return app.resolve(event, context)
        ```
        """
        self._router_middlewares = self._router_middlewares + middlewares

    def get(
        self,
        rule: str,
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: str = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        """Get route decorator with GET `method`

        Examples
        --------
        Simple example with a custom lambda handler using the Tracer capture_lambda_handler decorator

        ```python
        from aws_lambda_powertools import Tracer
        from aws_lambda_powertools.event_handler import APIGatewayRestResolver

        tracer = Tracer()
        app = APIGatewayRestResolver()

        @app.get("/get-call")
        def simple_get():
            return {"message": "Foo"}

        @tracer.capture_lambda_handler
        def lambda_handler(event, context):
            return app.resolve(event, context)
        ```
        """
        return self.route(
            rule,
            "GET",
            cors,
            compress,
            cache_control,
            summary,
            description,
            responses,
            response_description,
            tags,
            operation_id,
            include_in_schema,
            middlewares,
        )

    def post(
        self,
        rule: str,
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: str = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        """Post route decorator with POST `method`

        Examples
        --------
        Simple example with a custom lambda handler using the Tracer capture_lambda_handler decorator

        ```python
        from aws_lambda_powertools import Tracer
        from aws_lambda_powertools.event_handler import APIGatewayRestResolver

        tracer = Tracer()
        app = APIGatewayRestResolver()

        @app.post("/post-call")
        def simple_post():
            post_data: dict = app.current_event.json_body
            return {"message": post_data["value"]}

        @tracer.capture_lambda_handler
        def lambda_handler(event, context):
            return app.resolve(event, context)
        ```
        """
        return self.route(
            rule,
            "POST",
            cors,
            compress,
            cache_control,
            summary,
            description,
            responses,
            response_description,
            tags,
            operation_id,
            include_in_schema,
            middlewares,
        )

    def put(
        self,
        rule: str,
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: str = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        """Put route decorator with PUT `method`

        Examples
        --------
        Simple example with a custom lambda handler using the Tracer capture_lambda_handler decorator

        ```python
        from aws_lambda_powertools import Tracer
        from aws_lambda_powertools.event_handler import APIGatewayRestResolver

        tracer = Tracer()
        app = APIGatewayRestResolver()

        @app.put("/put-call")
        def simple_put():
            put_data: dict = app.current_event.json_body
            return {"message": put_data["value"]}

        @tracer.capture_lambda_handler
        def lambda_handler(event, context):
            return app.resolve(event, context)
        ```
        """
        return self.route(
            rule,
            "PUT",
            cors,
            compress,
            cache_control,
            summary,
            description,
            responses,
            response_description,
            tags,
            operation_id,
            include_in_schema,
            middlewares,
        )

    def delete(
        self,
        rule: str,
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: str = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        """Delete route decorator with DELETE `method`

        Examples
        --------
        Simple example with a custom lambda handler using the Tracer capture_lambda_handler decorator

        ```python
        from aws_lambda_powertools import Tracer
        from aws_lambda_powertools.event_handler import APIGatewayRestResolver

        tracer = Tracer()
        app = APIGatewayRestResolver()

        @app.delete("/delete-call")
        def simple_delete():
            return {"message": "deleted"}

        @tracer.capture_lambda_handler
        def lambda_handler(event, context):
            return app.resolve(event, context)
        ```
        """
        return self.route(
            rule,
            "DELETE",
            cors,
            compress,
            cache_control,
            summary,
            description,
            responses,
            response_description,
            tags,
            operation_id,
            include_in_schema,
            middlewares,
        )

    def patch(
        self,
        rule: str,
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: str = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable]] = None,
    ):
        """Patch route decorator with PATCH `method`

        Examples
        --------
        Simple example with a custom lambda handler using the Tracer capture_lambda_handler decorator

        ```python
        from aws_lambda_powertools import Tracer
        from aws_lambda_powertools.event_handler import APIGatewayRestResolver

        tracer = Tracer()
        app = APIGatewayRestResolver()

        @app.patch("/patch-call")
        def simple_patch():
            patch_data: dict = app.current_event.json_body
            patch_data["value"] = patched

            return {"message": patch_data}

        @tracer.capture_lambda_handler
        def lambda_handler(event, context):
            return app.resolve(event, context)
        ```
        """
        return self.route(
            rule,
            "PATCH",
            cors,
            compress,
            cache_control,
            summary,
            description,
            responses,
            response_description,
            tags,
            operation_id,
            include_in_schema,
            middlewares,
        )

    def _push_processed_stack_frame(self, frame: str):
        """
        Add Current Middleware to the Middleware Stack Frames
        The stack frames will be used when exceptions are thrown and Powertools
        debug is enabled by developers.
        """
        self.processed_stack_frames.append(frame)

    def _reset_processed_stack(self):
        """Reset the Processed Stack Frames"""
        self.processed_stack_frames.clear()

    def append_context(self, **additional_context):
        """Append key=value data as routing context"""
        self.context.update(**additional_context)

    def clear_context(self):
        """Resets routing context"""
        self.context.clear()


class MiddlewareFrame:
    """
    Creates a Middle Stack Wrapper instance to be used as a "Frame" in the overall stack of
    middleware functions.  Each instance contains the current middleware and the next
    middleware function to be called in the stack.

    In this way the middleware stack is constructed in a recursive fashion, with each middleware
    calling the next as a simple function call.  The actual Python call-stack will contain
    each MiddlewareStackWrapper "Frame", meaning any Middleware function can cause the
    entire Middleware call chain to be exited early (short-circuited) by raising an exception
    or by simply returning early with a custom Response.  The decision to short-circuit the middleware
    chain is at the user's discretion but instantly available due to the Wrapped nature of the
    callable constructs in the Middleware stack and each Middleware function having complete control over
    whether the "Next" handler in the stack is called or not.

    Parameters
    ----------
    current_middleware : Callable
        The current middleware function to be called as a request is processed.
    next_middleware : Callable
        The next middleware in the middleware stack.
    """

    def __init__(
        self,
        current_middleware: Callable[..., Any],
        next_middleware: Callable[..., Any],
    ) -> None:
        self.current_middleware: Callable[..., Any] = current_middleware
        self.next_middleware: Callable[..., Any] = next_middleware
        self._next_middleware_name = next_middleware.__name__

    @property
    def __name__(self) -> str:  # noqa: A003
        """Current middleware name

        It ensures backward compatibility with view functions being callable. This
        improves debugging since we need both current and next middlewares/callable names.
        """
        return self.current_middleware.__name__

    def __str__(self) -> str:
        """Identify current middleware identity and call chain for debugging purposes."""
        middleware_name = self.__name__
        return f"[{middleware_name}] next call chain is {middleware_name} -> {self._next_middleware_name}"

    def __call__(self, app: "ApiGatewayResolver") -> Union[Dict, Tuple, Response]:
        """
        Call the middleware Frame to process the request.

        Parameters
        ----------
        app: BaseRouter
            The router instance

        Returns
        -------
        Union[Dict, Tuple, Response]
            (tech-debt for backward compatibility).  The response type should be a
            Response object in all cases excepting when the original API route handler
            is called which will return one of 3 outputs.

        """
        # Do debug printing and push processed stack frame AFTER calling middleware
        # else the stack frame text of `current calling next` is confusing.
        logger.debug("MiddlewareFrame: %s", self)
        app._push_processed_stack_frame(str(self))

        return self.current_middleware(app, self.next_middleware)


def _registered_api_adapter(
    app: "ApiGatewayResolver",
    next_middleware: Callable[..., Any],
) -> Union[Dict, Tuple, Response]:
    """
    Calls the registered API using the "_route_args" from the Resolver context to ensure the last call
    in the chain will match the API route function signature and ensure that Powertools passes the API
    route handler the expected arguments.

    **IMPORTANT: This internal middleware ensures the actual API route is called with the correct call signature
    and it MUST be the final frame in the middleware stack.  This can only be removed when the API Route
    function accepts `app: BaseRouter` as the first argument - which is the breaking change.

    Parameters
    ----------
    app: ApiGatewayResolver
        The API Gateway resolver
    next_middleware: Callable[..., Any]
        The function to handle the API

    Returns
    -------
    Response
        The API Response Object

    """
    route_args: Dict = app.context.get("_route_args", {})
    logger.debug(f"Calling API Route Handler: {route_args}")

    return app._to_response(next_middleware(**route_args))


class ApiGatewayResolver(BaseRouter):
    """API Gateway and ALB proxy resolver

    Examples
    --------
    Simple example with a custom lambda handler using the Tracer capture_lambda_handler decorator

    ```python
    from aws_lambda_powertools import Tracer
    from aws_lambda_powertools.event_handler import APIGatewayRestResolver

    tracer = Tracer()
    app = APIGatewayRestResolver()

    @app.get("/get-call")
    def simple_get():
        return {"message": "Foo"}

    @app.post("/post-call")
    def simple_post():
        post_data: dict = app.current_event.json_body
        return {"message": post_data["value"]}

    @tracer.capture_lambda_handler
    def lambda_handler(event, context):
        return app.resolve(event, context)
    ```
    """

    def __init__(
        self,
        proxy_type: Enum = ProxyEventType.APIGatewayProxyEvent,
        cors: Optional[CORSConfig] = None,
        debug: Optional[bool] = None,
        serializer: Optional[Callable[[Dict], str]] = None,
        strip_prefixes: Optional[List[Union[str, Pattern]]] = None,
        enable_validation: bool = False,
    ):
        """
        Parameters
        ----------
        proxy_type: ProxyEventType
            Proxy request type, defaults to API Gateway V1
        cors: CORSConfig
            Optionally configure and enabled CORS. Not each route will need to have to cors=True
        debug: Optional[bool]
            Enables debug mode, by default False. Can be also be enabled by "POWERTOOLS_DEV"
            environment variable
        serializer: Callable, optional
            function to serialize `obj` to a JSON formatted `str`, by default json.dumps
        strip_prefixes: List[Union[str, Pattern]], optional
            optional list of prefixes to be removed from the request path before doing the routing.
            This is often used with api gateways with multiple custom mappings.
            Each prefix can be a static string or a compiled regex pattern
        enable_validation: Optional[bool]
            Enables validation of the request body against the route schema, by default False.
        """
        self._proxy_type = proxy_type
        self._dynamic_routes: List[Route] = []
        self._static_routes: List[Route] = []
        self._route_keys: List[str] = []
        self._exception_handlers: Dict[Type, Callable] = {}
        self._cors = cors
        self._cors_enabled: bool = cors is not None
        self._cors_methods: Set[str] = {"OPTIONS"}
        self._debug = self._has_debug(debug)
        self._enable_validation = enable_validation
        self._strip_prefixes = strip_prefixes
        self.context: Dict = {}  # early init as customers might add context before event resolution
        self.processed_stack_frames = []
        self._response_builder_class = ResponseBuilder[BaseProxyEvent]

        # Allow for a custom serializer or a concise json serialization
        self._serializer = serializer or partial(json.dumps, separators=(",", ":"), cls=Encoder)

        if self._enable_validation:
            from aws_lambda_powertools.event_handler.middlewares.openapi_validation import OpenAPIValidationMiddleware

            self.use([OpenAPIValidationMiddleware()])

    def get_openapi_schema(
        self,
        *,
        title: str = "Powertools API",
        version: str = DEFAULT_API_VERSION,
        openapi_version: str = DEFAULT_OPENAPI_VERSION,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        servers: Optional[List["Server"]] = None,
        terms_of_service: Optional[str] = None,
        contact: Optional["Contact"] = None,
        license_info: Optional["License"] = None,
    ) -> "OpenAPI":
        """
        Returns the OpenAPI schema as a pydantic model.

        Parameters
        ----------
        title: str
            The title of the application.
        version: str
            The version of the OpenAPI document (which is distinct from the OpenAPI Specification version or the API
        openapi_version: str, default = "3.0.0"
            The version of the OpenAPI Specification (which the document uses).
        summary: str, optional
            A short summary of what the application does.
        description: str, optional
            A verbose explanation of the application behavior.
        tags: List[str], optional
            A list of tags used by the specification with additional metadata.
        servers: List[Server], optional
            An array of Server Objects, which provide connectivity information to a target server.
        terms_of_service: str, optional
            A URL to the Terms of Service for the API. MUST be in the format of a URL.
        contact: Contact, optional
            The contact information for the exposed API.
        license_info: License, optional
            The license information for the exposed API.

        Returns
        -------
        OpenAPI: pydantic model
            The OpenAPI schema as a pydantic model.
        """

        from aws_lambda_powertools.event_handler.openapi.compat import (
            GenerateJsonSchema,
            get_compat_model_name_map,
            get_definitions,
        )
        from aws_lambda_powertools.event_handler.openapi.models import OpenAPI, PathItem, Server
        from aws_lambda_powertools.event_handler.openapi.types import (
            COMPONENT_REF_TEMPLATE,
        )

        # Start with the bare minimum required for a valid OpenAPI schema
        info: Dict[str, Any] = {"title": title, "version": version}

        optional_fields = {
            "summary": summary,
            "description": description,
            "termsOfService": terms_of_service,
            "contact": contact,
            "license": license_info,
        }

        info.update({field: value for field, value in optional_fields.items() if value})

        output: Dict[str, Any] = {"openapi": openapi_version, "info": info}
        if servers:
            output["servers"] = servers
        else:
            # If the servers property is not provided, or is an empty array, the default value would be a Server Object
            # with an url value of /.
            output["servers"] = [Server(url="/")]

        components: Dict[str, Dict[str, Any]] = {}
        paths: Dict[str, Dict[str, Any]] = {}
        operation_ids: Set[str] = set()

        all_routes = self._dynamic_routes + self._static_routes
        all_fields = self._get_fields_from_routes(all_routes)
        model_name_map = get_compat_model_name_map(all_fields)

        # Collect all models and definitions
        schema_generator = GenerateJsonSchema(ref_template=COMPONENT_REF_TEMPLATE)
        field_mapping, definitions = get_definitions(
            fields=all_fields,
            schema_generator=schema_generator,
            model_name_map=model_name_map,
        )

        # Add routes to the OpenAPI schema
        for route in all_routes:
            if not route.include_in_schema:
                continue

            result = route._get_openapi_path(
                dependant=route.dependant,
                operation_ids=operation_ids,
                model_name_map=model_name_map,
                field_mapping=field_mapping,
            )
            if result:
                path, path_definitions = result
                if path:
                    paths.setdefault(route.openapi_path, {}).update(path)
                if path_definitions:
                    definitions.update(path_definitions)

        if definitions:
            components["schemas"] = {k: definitions[k] for k in sorted(definitions)}
        if components:
            output["components"] = components
        if tags:
            output["tags"] = [{"name": tag} for tag in tags]

        output["paths"] = {k: PathItem(**v) for k, v in paths.items()}

        return OpenAPI(**output)

    def get_openapi_json_schema(
        self,
        *,
        title: str = "Powertools API",
        version: str = DEFAULT_API_VERSION,
        openapi_version: str = DEFAULT_OPENAPI_VERSION,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        servers: Optional[List["Server"]] = None,
        terms_of_service: Optional[str] = None,
        contact: Optional["Contact"] = None,
        license_info: Optional["License"] = None,
    ) -> str:
        """
        Returns the OpenAPI schema as a JSON serializable dict

        Parameters
        ----------
        title: str
            The title of the application.
        version: str
            The version of the OpenAPI document (which is distinct from the OpenAPI Specification version or the API
        openapi_version: str, default = "3.0.0"
            The version of the OpenAPI Specification (which the document uses).
        summary: str, optional
            A short summary of what the application does.
        description: str, optional
            A verbose explanation of the application behavior.
        tags: List[str], optional
            A list of tags used by the specification with additional metadata.
        servers: List[Server], optional
            An array of Server Objects, which provide connectivity information to a target server.
        terms_of_service: str, optional
            A URL to the Terms of Service for the API. MUST be in the format of a URL.
        contact: Contact, optional
            The contact information for the exposed API.
        license_info: License, optional
            The license information for the exposed API.

        Returns
        -------
        str
            The OpenAPI schema as a JSON serializable dict.
        """
        from aws_lambda_powertools.event_handler.openapi.compat import model_json

        return model_json(
            self.get_openapi_schema(
                title=title,
                version=version,
                openapi_version=openapi_version,
                summary=summary,
                description=description,
                tags=tags,
                servers=servers,
                terms_of_service=terms_of_service,
                contact=contact,
                license_info=license_info,
            ),
            by_alias=True,
            exclude_none=True,
            indent=2,
        )

    def enable_swagger(
        self,
        *,
        path: str = "/swagger",
        title: str = "Powertools for AWS Lambda (Python) API",
        version: str = DEFAULT_API_VERSION,
        openapi_version: str = DEFAULT_OPENAPI_VERSION,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        servers: Optional[List["Server"]] = None,
        terms_of_service: Optional[str] = None,
        contact: Optional["Contact"] = None,
        license_info: Optional["License"] = None,
        swagger_base_url: Optional[str] = None,
        middlewares: Optional[List[Callable[..., Response]]] = None,
    ):
        """
        Returns the OpenAPI schema as a JSON serializable dict

        Parameters
        ----------
        path: str, default = "/swagger"
            The path to the swagger UI.
        title: str
            The title of the application.
        version: str
            The version of the OpenAPI document (which is distinct from the OpenAPI Specification version or the API
        openapi_version: str, default = "3.0.0"
            The version of the OpenAPI Specification (which the document uses).
        summary: str, optional
            A short summary of what the application does.
        description: str, optional
            A verbose explanation of the application behavior.
        tags: List[str], optional
            A list of tags used by the specification with additional metadata.
        servers: List[Server], optional
            An array of Server Objects, which provide connectivity information to a target server.
        terms_of_service: str, optional
            A URL to the Terms of Service for the API. MUST be in the format of a URL.
        contact: Contact, optional
            The contact information for the exposed API.
        license_info: License, optional
            The license information for the exposed API.
        swagger_base_url: str, optional
            The base url for the swagger UI. If not provided, we will serve a recent version of the Swagger UI.
        middlewares: List[Callable[..., Response]], optional
            List of middlewares to be used for the swagger route.
        """
        from aws_lambda_powertools.event_handler.openapi.models import Server

        if not swagger_base_url:

            @self.get("/swagger.js", include_in_schema=False)
            def swagger_js():
                body = Path.open(Path(__file__).parent / "openapi" / "swagger_ui" / "swagger-ui-bundle.min.js").read()
                return Response(
                    status_code=200,
                    content_type="text/javascript",
                    body=body,
                )

            @self.get("/swagger.css", include_in_schema=False)
            def swagger_css():
                body = Path.open(Path(__file__).parent / "openapi" / "swagger_ui" / "swagger-ui.min.css").read()
                return Response(
                    status_code=200,
                    content_type="text/css",
                    body=body,
                )

        @self.get(path, middlewares=middlewares, include_in_schema=False)
        def swagger_handler():
            base_path = self._get_base_path()

            if swagger_base_url:
                swagger_js = f"{swagger_base_url}/swagger-ui-bundle.min.js"
                swagger_css = f"{swagger_base_url}/swagger-ui.min.css"
            else:
                swagger_js = f"{base_path}/swagger.js"
                swagger_css = f"{base_path}/swagger.css"

            openapi_servers = servers or [Server(url=(base_path or "/"))]

            spec = self.get_openapi_json_schema(
                title=title,
                version=version,
                openapi_version=openapi_version,
                summary=summary,
                description=description,
                tags=tags,
                servers=openapi_servers,
                terms_of_service=terms_of_service,
                contact=contact,
                license_info=license_info,
            )

            body = generate_swagger_html(spec, swagger_js, swagger_css)

            return Response(
                status_code=200,
                content_type="text/html",
                body=body,
            )

    def route(
        self,
        rule: str,
        method: Union[str, Union[List[str], Tuple[str]]],
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: str = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        """Route decorator includes parameter `method`"""

        def register_resolver(func: Callable):
            methods = (method,) if isinstance(method, str) else method
            logger.debug(f"Adding route using rule {rule} and methods: {','.join((m.upper() for m in methods))}")

            cors_enabled = self._cors_enabled if cors is None else cors

            for item in methods:
                _route = Route(
                    item,
                    rule,
                    self._compile_regex(rule),
                    func,
                    cors_enabled,
                    compress,
                    cache_control,
                    summary,
                    description,
                    responses,
                    response_description,
                    tags,
                    operation_id,
                    include_in_schema,
                    middlewares,
                )

                # The more specific route wins.
                # We store dynamic (/studies/{studyid}) and static routes (/studies/fetch) separately.
                # Then attempt a match for static routes before dynamic routes.
                # This ensures that the most specific route is prioritized and processed first (studies/fetch).
                if _route.rule.groups > 0:
                    self._dynamic_routes.append(_route)
                else:
                    self._static_routes.append(_route)

                self._create_route_key(item, rule)

                if cors_enabled:
                    logger.debug(f"Registering method {item.upper()} to Allow Methods in CORS")
                    self._cors_methods.add(item.upper())

            return func

        return register_resolver

    def resolve(self, event, context) -> Dict[str, Any]:
        """Resolves the response based on the provide event and decorator routes

        Parameters
        ----------
        event: Dict[str, Any]
            Event
        context: LambdaContext
            Lambda context
        Returns
        -------
        dict
            Returns the dict response
        """
        if isinstance(event, BaseProxyEvent):
            warnings.warn(
                "You don't need to serialize event to Event Source Data Class when using Event Handler; "
                "see issue #1152",
                stacklevel=2,
            )
            event = event.raw_event

        if self._debug:
            print(self._serializer(event))

        # Populate router(s) dependencies without keeping a reference to each registered router
        BaseRouter.current_event = self._to_proxy_event(event)
        BaseRouter.lambda_context = context

        response = self._resolve().build(self.current_event, self._cors)

        # Debug print Processed Middlewares
        if self._debug:
            print("\nProcessed Middlewares:")
            print("======================")
            print("\n".join(self.processed_stack_frames))
            print("======================")

        self.clear_context()

        return response

    def __call__(self, event, context) -> Any:
        return self.resolve(event, context)

    def _create_route_key(self, item: str, rule: str):
        route_key = item + rule
        if route_key in self._route_keys:
            warnings.warn(
                f"A route like this was already registered. method: '{item}' rule: '{rule}'",
                stacklevel=2,
            )
        self._route_keys.append(route_key)

    def _get_base_path(self) -> str:
        raise NotImplementedError()

    @staticmethod
    def _has_debug(debug: Optional[bool] = None) -> bool:
        # It might have been explicitly switched off (debug=False)
        if debug is not None:
            return debug

        return powertools_dev_is_set()

    @staticmethod
    def _compile_regex(rule: str, base_regex: str = _ROUTE_REGEX):
        """Precompile regex pattern

        Logic
        -----

        1. Find any dynamic routes defined as <pattern>
            e.g. @app.get("/accounts/<account_id>")
        2. Create a new regex by substituting every dynamic route found as a named group (?P<group>),
        and match whole words only (word boundary) instead of a greedy match

            non-greedy example with word boundary

                rule: '/accounts/<account_id>'
                regex: r'/accounts/(?P<account_id>\\w+\\b)'

                value: /accounts/123/some_other_path
                account_id: 123

            greedy example without word boundary

                regex: r'/accounts/(?P<account_id>.+)'

                value: /accounts/123/some_other_path
                account_id: 123/some_other_path
        3. Compiles a regex and include start (^) and end ($) in between for an exact match

        NOTE: See #520 for context
        """
        rule_regex: str = re.sub(_DYNAMIC_ROUTE_PATTERN, _NAMED_GROUP_BOUNDARY_PATTERN, rule)
        return re.compile(base_regex.format(rule_regex))

    def _to_proxy_event(self, event: Dict) -> BaseProxyEvent:  # noqa: PLR0911  # ignore many returns
        """Convert the event dict to the corresponding data class"""
        if self._proxy_type == ProxyEventType.APIGatewayProxyEvent:
            logger.debug("Converting event to API Gateway REST API contract")
            return APIGatewayProxyEvent(event)
        if self._proxy_type == ProxyEventType.APIGatewayProxyEventV2:
            logger.debug("Converting event to API Gateway HTTP API contract")
            return APIGatewayProxyEventV2(event)
        if self._proxy_type == ProxyEventType.BedrockAgentEvent:
            logger.debug("Converting event to Bedrock Agent contract")
            return BedrockAgentEvent(event)
        if self._proxy_type == ProxyEventType.LambdaFunctionUrlEvent:
            logger.debug("Converting event to Lambda Function URL contract")
            return LambdaFunctionUrlEvent(event)
        if self._proxy_type == ProxyEventType.VPCLatticeEvent:
            logger.debug("Converting event to VPC Lattice contract")
            return VPCLatticeEvent(event)
        if self._proxy_type == ProxyEventType.VPCLatticeEventV2:
            logger.debug("Converting event to VPC LatticeV2 contract")
            return VPCLatticeEventV2(event)
        logger.debug("Converting event to ALB contract")
        return ALBEvent(event)

    def _resolve(self) -> ResponseBuilder:
        """Resolves the response or return the not found response"""
        method = self.current_event.http_method.upper()
        path = self._remove_prefix(self.current_event.path)

        for route in self._static_routes + self._dynamic_routes:
            if method != route.method:
                continue
            match_results: Optional[Match] = route.rule.match(path)
            if match_results:
                logger.debug("Found a registered route. Calling function")
                # Add matched Route reference into the Resolver context
                self.append_context(_route=route, _path=path)

                route_keys = self._convert_matches_into_route_keys(match_results)
                return self._call_route(route, route_keys)  # pass fn args

        logger.debug(f"No match found for path {path} and method {method}")
        return self._not_found(method)

    def _remove_prefix(self, path: str) -> str:
        """Remove the configured prefix from the path"""
        if not isinstance(self._strip_prefixes, list):
            return path

        for prefix in self._strip_prefixes:
            if isinstance(prefix, str):
                if path == prefix:
                    return "/"

                if self._path_starts_with(path, prefix):
                    return path[len(prefix) :]

            if isinstance(prefix, Pattern):
                path = re.sub(prefix, "", path)

                # When using regexes, we might get into a point where everything is removed
                # from the string, so we check if it's empty and return /, since there's nothing
                # else to strip anymore.
                if not path:
                    return "/"

        return path

    def _convert_matches_into_route_keys(self, match: Match) -> Dict[str, str]:
        """Converts the regex match into a dict of route keys"""
        return match.groupdict()

    @staticmethod
    def _path_starts_with(path: str, prefix: str):
        """Returns true if the `path` starts with a prefix plus a `/`"""
        if not isinstance(prefix, str) or prefix == "":
            return False

        return path.startswith(prefix + "/")

    def _not_found(self, method: str) -> ResponseBuilder:
        """Called when no matching route was found and includes support for the cors preflight response"""
        headers: Dict[str, Union[str, List[str]]] = {}
        if self._cors:
            logger.debug("CORS is enabled, updating headers.")
            headers.update(self._cors.to_dict(self.current_event.get_header_value("Origin")))

            if method == "OPTIONS":
                logger.debug("Pre-flight request detected. Returning CORS with null response")
                headers["Access-Control-Allow-Methods"] = ",".join(sorted(self._cors_methods))
                return ResponseBuilder(
                    response=Response(status_code=204, content_type=None, headers=headers, body=""),
                    serializer=self._serializer,
                )

        handler = self._lookup_exception_handler(NotFoundError)
        if handler:
            return self._response_builder_class(response=handler(NotFoundError()), serializer=self._serializer)

        return self._response_builder_class(
            response=Response(
                status_code=HTTPStatus.NOT_FOUND.value,
                content_type=content_types.APPLICATION_JSON,
                headers=headers,
                body={"statusCode": HTTPStatus.NOT_FOUND.value, "message": "Not found"},
            ),
            serializer=self._serializer,
        )

    def _call_route(self, route: Route, route_arguments: Dict[str, str]) -> ResponseBuilder:
        """Actually call the matching route with any provided keyword arguments."""
        try:
            # Reset Processed stack for Middleware (for debugging purposes)
            self._reset_processed_stack()

            return self._response_builder_class(
                response=self._to_response(
                    route(router_middlewares=self._router_middlewares, app=self, route_arguments=route_arguments),
                ),
                serializer=self._serializer,
                route=route,
            )
        except Exception as exc:
            # If exception is handled then return the response builder to reduce noise
            response_builder = self._call_exception_handler(exc, route)
            if response_builder:
                return response_builder

            logger.exception(exc)
            if self._debug:
                # If the user has turned on debug mode,
                # we'll let the original exception propagate, so
                # they get more information about what went wrong.
                return self._response_builder_class(
                    response=Response(
                        status_code=500,
                        content_type=content_types.TEXT_PLAIN,
                        body="".join(traceback.format_exc()),
                    ),
                    serializer=self._serializer,
                    route=route,
                )

            raise

    def not_found(self, func: Optional[Callable] = None):
        if func is None:
            return self.exception_handler(NotFoundError)
        return self.exception_handler(NotFoundError)(func)

    def exception_handler(self, exc_class: Union[Type[Exception], List[Type[Exception]]]):
        def register_exception_handler(func: Callable):
            if isinstance(exc_class, list):
                for exp in exc_class:
                    self._exception_handlers[exp] = func
            else:
                self._exception_handlers[exc_class] = func
            return func

        return register_exception_handler

    def _lookup_exception_handler(self, exp_type: Type) -> Optional[Callable]:
        # Use "Method Resolution Order" to allow for matching against a base class
        # of an exception
        for cls in exp_type.__mro__:
            if cls in self._exception_handlers:
                return self._exception_handlers[cls]
        return None

    def _call_exception_handler(self, exp: Exception, route: Route) -> Optional[ResponseBuilder]:
        handler = self._lookup_exception_handler(type(exp))
        if handler:
            try:
                return self._response_builder_class(response=handler(exp), serializer=self._serializer, route=route)
            except ServiceError as service_error:
                exp = service_error

        if isinstance(exp, RequestValidationError):
            # For security reasons, we hide msg details (don't leak Python, Pydantic or file names)
            errors = [{"loc": e["loc"], "type": e["type"]} for e in exp.errors()]

            return self._response_builder_class(
                response=Response(
                    status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                    content_type=content_types.APPLICATION_JSON,
                    body={"statusCode": HTTPStatus.UNPROCESSABLE_ENTITY, "detail": errors},
                ),
                serializer=self._serializer,
                route=route,
            )

        if isinstance(exp, ServiceError):
            return self._response_builder_class(
                response=Response(
                    status_code=exp.status_code,
                    content_type=content_types.APPLICATION_JSON,
                    body={"statusCode": exp.status_code, "message": exp.msg},
                ),
                serializer=self._serializer,
                route=route,
            )

        return None

    def _to_response(self, result: Union[Dict, Tuple, Response]) -> Response:
        """Convert the route's result to a Response

         3 main result types are supported:

        - Dict[str, Any]: Rest api response with just the Dict to json stringify and content-type is set to
          application/json
        - Tuple[dict, int]: Same dict handling as above but with the option of including a status code
        - Response: returned as is, and allows for more flexibility
        """
        status_code = HTTPStatus.OK
        if isinstance(result, Response):
            return result
        elif isinstance(result, tuple) and len(result) == 2:
            # Unpack result dict and status code from tuple
            result, status_code = result

        logger.debug("Simple response detected, serializing return before constructing final response")
        return Response(
            status_code=status_code,
            content_type=content_types.APPLICATION_JSON,
            body=result,
        )

    def include_router(self, router: "Router", prefix: Optional[str] = None) -> None:
        """Adds all routes and context defined in a router

        Parameters
        ----------
        router : Router
            The Router containing a list of routes to be registered after the existing routes
        prefix : str, optional
            An optional prefix to be added to the originally defined rule
        """

        # Add reference to parent ApiGatewayResolver to support use cases where people subclass it to add custom logic
        router.api_resolver = self

        logger.debug("Merging App context with Router context")
        self.context.update(**router.context)

        logger.debug("Appending Router middlewares into App middlewares.")
        self._router_middlewares = self._router_middlewares + router._router_middlewares

        # use pointer to allow context clearance after event is processed e.g., resolve(evt, ctx)
        router.context = self.context

        # Iterate through the routes defined in the router to configure and apply middlewares for each route
        for route, func in router._routes.items():
            new_route = route

            if prefix:
                rule = route[0]
                rule = prefix if rule == "/" else f"{prefix}{rule}"
                new_route = (rule, *route[1:])

            # Middlewares are stored by route separately - must grab them to include
            # Middleware store the route without prefix, so we must not include prefix when grabbing
            middlewares = router._routes_with_middleware.get(route)

            # Need to use "type: ignore" here since mypy does not like a named parameter after
            # tuple expansion since may cause duplicate named parameters in the function signature.
            # In this case this is not possible since the tuple expansion is from a hashable source
            # and the `middlewares` List is a non-hashable structure so will never be included.
            # Still need to ignore for mypy checks or will cause failures (false-positive)
            self.route(*new_route, middlewares=middlewares)(func)  # type: ignore

    @staticmethod
    def _get_fields_from_routes(routes: Sequence[Route]) -> List["ModelField"]:
        """
        Returns a list of fields from the routes
        """

        from aws_lambda_powertools.event_handler.openapi.compat import ModelField
        from aws_lambda_powertools.event_handler.openapi.dependant import (
            get_flat_params,
        )

        body_fields_from_routes: List["ModelField"] = []
        responses_from_routes: List["ModelField"] = []
        request_fields_from_routes: List["ModelField"] = []

        for route in routes:
            if route.body_field:
                if not isinstance(route.body_field, ModelField):
                    raise AssertionError("A request body myst be a Pydantic Field")
                body_fields_from_routes.append(route.body_field)

            params = get_flat_params(route.dependant)
            request_fields_from_routes.extend(params)

            if route.dependant.return_param:
                responses_from_routes.append(route.dependant.return_param)

        flat_models = list(responses_from_routes + request_fields_from_routes + body_fields_from_routes)
        return flat_models


class Router(BaseRouter):
    """Router helper class to allow splitting ApiGatewayResolver into multiple files"""

    def __init__(self):
        self._routes: Dict[tuple, Callable] = {}
        self._routes_with_middleware: Dict[tuple, List[Callable]] = {}
        self.api_resolver: Optional[BaseRouter] = None
        self.context = {}  # early init as customers might add context before event resolution

    def route(
        self,
        rule: str,
        method: Union[str, Union[List[str], Tuple[str]]],
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: Optional[str] = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        def register_route(func: Callable):
            # All dict keys needs to be hashable. So we'll need to do some conversions:
            methods = (method,) if isinstance(method, str) else tuple(method)
            frozen_responses = _FrozenDict(responses) if responses else None
            frozen_tags = frozenset(tags) if tags else None

            route_key = (
                rule,
                methods,
                cors,
                compress,
                cache_control,
                summary,
                description,
                frozen_responses,
                response_description,
                frozen_tags,
                operation_id,
                include_in_schema,
            )

            # Collate Middleware for routes
            if middlewares is not None:
                for handler in middlewares:
                    if self._routes_with_middleware.get(route_key) is None:
                        self._routes_with_middleware[route_key] = [handler]
                    else:
                        self._routes_with_middleware[route_key].append(handler)
            else:
                self._routes_with_middleware[route_key] = []

            self._routes[route_key] = func

            return func

        return register_route


class APIGatewayRestResolver(ApiGatewayResolver):
    current_event: APIGatewayProxyEvent

    def __init__(
        self,
        cors: Optional[CORSConfig] = None,
        debug: Optional[bool] = None,
        serializer: Optional[Callable[[Dict], str]] = None,
        strip_prefixes: Optional[List[Union[str, Pattern]]] = None,
        enable_validation: bool = False,
    ):
        """Amazon API Gateway REST and HTTP API v1 payload resolver"""
        super().__init__(
            ProxyEventType.APIGatewayProxyEvent,
            cors,
            debug,
            serializer,
            strip_prefixes,
            enable_validation,
        )

    def _get_base_path(self) -> str:
        # 3 different scenarios:
        #
        # 1. SAM local: even though a stage variable is sent to the Lambda function, it's not used in the path
        # 2. API Gateway REST API: stage variable is used in the path
        # 3. API Gateway REST Custom Domain: stage variable is not used in the path
        #
        # To solve the 3 scenarios, we try to match the beginning of the path with the stage variable
        stage = self.current_event.request_context.stage
        if stage and stage != "$default" and self.current_event.request_context.path.startswith(f"/{stage}"):
            return f"/{stage}"
        return ""

    # override route to ignore trailing "/" in routes for REST API
    def route(
        self,
        rule: str,
        method: Union[str, Union[List[str], Tuple[str]]],
        cors: Optional[bool] = None,
        compress: bool = False,
        cache_control: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        responses: Optional[Dict[int, Dict[str, Any]]] = None,
        response_description: str = _DEFAULT_OPENAPI_RESPONSE_DESCRIPTION,
        tags: Optional[List[str]] = None,
        operation_id: Optional[str] = None,
        include_in_schema: bool = True,
        middlewares: Optional[List[Callable[..., Any]]] = None,
    ):
        # NOTE: see #1552 for more context.
        return super().route(
            rule.rstrip("/"),
            method,
            cors,
            compress,
            cache_control,
            summary,
            description,
            responses,
            response_description,
            tags,
            operation_id,
            include_in_schema,
            middlewares,
        )

    # Override _compile_regex to exclude trailing slashes for route resolution
    @staticmethod
    def _compile_regex(rule: str, base_regex: str = _ROUTE_REGEX):
        return super(APIGatewayRestResolver, APIGatewayRestResolver)._compile_regex(rule, "^{}/*$")


class APIGatewayHttpResolver(ApiGatewayResolver):
    current_event: APIGatewayProxyEventV2

    def __init__(
        self,
        cors: Optional[CORSConfig] = None,
        debug: Optional[bool] = None,
        serializer: Optional[Callable[[Dict], str]] = None,
        strip_prefixes: Optional[List[Union[str, Pattern]]] = None,
        enable_validation: bool = False,
    ):
        """Amazon API Gateway HTTP API v2 payload resolver"""
        super().__init__(
            ProxyEventType.APIGatewayProxyEventV2,
            cors,
            debug,
            serializer,
            strip_prefixes,
            enable_validation,
        )

    def _get_base_path(self) -> str:
        # 3 different scenarios:
        #
        # 1. SAM local: even though a stage variable is sent to the Lambda function, it's not used in the path
        # 2. API Gateway HTTP API: stage variable is used in the path
        # 3. API Gateway HTTP Custom Domain: stage variable is not used in the path
        #
        # To solve the 3 scenarios, we try to match the beginning of the path with the stage variable
        stage = self.current_event.request_context.stage
        if stage and stage != "$default" and self.current_event.request_context.http.path.startswith(f"/{stage}"):
            return f"/{stage}"
        return ""


class ALBResolver(ApiGatewayResolver):
    current_event: ALBEvent

    def __init__(
        self,
        cors: Optional[CORSConfig] = None,
        debug: Optional[bool] = None,
        serializer: Optional[Callable[[Dict], str]] = None,
        strip_prefixes: Optional[List[Union[str, Pattern]]] = None,
        enable_validation: bool = False,
    ):
        """Amazon Application Load Balancer (ALB) resolver"""
        super().__init__(ProxyEventType.ALBEvent, cors, debug, serializer, strip_prefixes, enable_validation)

    def _get_base_path(self) -> str:
        # ALB doesn't have a stage variable, so we just return an empty string
        return ""



from __future__ import annotations

import secrets
import typing as t

from pydantic import BaseConfig, BaseModel, create_model, validator
from typing_extensions import Self

from datapane.client import DPClientError
from datapane.common.dp_types import StrEnum
from datapane.common.viewxml_utils import mk_attribs

from .base import BaseBlock, BlockId
from .parameters import E, Parameter

if t.TYPE_CHECKING:
    from datapane.common.viewxml_utils import ElementT
    from datapane.view import Blocks


def gen_name() -> str:
    """Return a (safe) name for use in a Block"""
    return f"id-{secrets.token_urlsafe(8)}"


class Empty(BaseBlock):
    """
    An empty block that can be updated / replaced later

    Args:
        name: A unique name for the block to reference when updating the report
    """

    _tag = "Empty"

    def __init__(self, name: BlockId):
        super().__init__(name=name)


class DummyModelConfig(BaseConfig):
    """Pydantic Model Config class for generated Controls models"""

    title = "Dynamic Parameters Model"
    anystr_strip_whitespace = True
    # use_enum_values = False
    # extra = 'ignore'
    allow_mutation = False
    frozen = True


class Controls:
    label: t.Optional[str]
    params: t.List[Parameter]

    def __init__(self, *params: Parameter, label: t.Optional[str] = None, **kwarg_params: Parameter):
        self.label = label
        self.params = self._make_params(*params, **kwarg_params)
        self._check()

    @classmethod
    def empty(cls) -> Controls:
        return cls()

    def __add__(self, other: Controls) -> Self:
        self.params.extend(other.params)
        self._check()
        return self

    def _make_params(self, *args: Parameter, **kwargs: Parameter) -> t.List[Parameter]:
        if args and kwargs:
            raise ValueError("Cannot have Parameters passed as both args and kwargs")
        if args:
            if any(not p.name for p in args):
                raise ValueError("Unnamed parameters cannot be passed as args")
            return list(args)
        params = []
        for name, param in kwargs.items():
            param.name = param.attribs["name"] = name
            params.append(param)
        return params

    def _check(self) -> None:
        # ensure all names are unique
        if len(self.params) != len({p.name for p in self.params}):
            raise ValueError("Several parameters have the same name")

    def _to_pydantic(self) -> t.Type[BaseModel]:
        """Return a dynamically generated Pydantic model for the controls"""
        fields = {p.name: p._as_field() for p in self.params}
        validators = {f"{p.name}_validator": validator(p.name, allow_reuse=True)(p._validator()) for p in self.params}
        DummyModel = create_model("DummyModel", **fields, __config__=DummyModelConfig, __validators__=validators)
        return DummyModel

    def _to_xml(self) -> ElementT:
        attribs = mk_attribs(label=self.label)
        return E.Controls(*[p._to_xml() for p in self.params], **attribs)

    @property
    def is_cacheable(self) -> bool:
        """Can the controls be cached"""
        return all(p.cacheable for p in self.params)

    @property
    def param_names(self) -> t.Set[str]:
        return {p.name for p in self.params}


class Swap(StrEnum):
    REPLACE = "replace"
    INNER = "inner"
    PREPEND = "prepend"
    APPEND = "append"


class Trigger(StrEnum):
    SUBMIT = "submit"
    SCHEDULE = "schedule"  # every 30s
    VISIBLE = "visible"
    LOAD = "load"
    MOUNT = "mount"


class TargetMode(StrEnum):
    SELF = "self"
    BELOW = "below"
    SIDE = "side"
    # TODO - 'parent' - find the nearest group ancestor and replace there
    # PARENT = "parent"


TargetT = t.Union[BlockId, TargetMode, BaseBlock]
ControlsT = t.Union[Controls, t.Iterable[Parameter], t.Mapping[str, Parameter], None]
FunctionT = t.Callable[..., "Blocks"]


class Compute(BaseBlock):
    """Compute blocks allow for dynamic views based on running python functions that return blocks"""

    _tag = "Compute"

    def __init__(
        self,
        function: FunctionT,
        target: TargetT,
        controls: ControlsT = None,
        name: BlockId = None,
        label: t.Optional[str] = None,
        swap: Swap = Swap.REPLACE,
        trigger: Trigger = Trigger.SUBMIT,
        immediate: bool = False,
        # function params - wrap around @dp.function
        cache: bool = False,
        submit_label: str = "Go",
        timer: int = 30,
    ):
        """
        Create a compute block that runs the provided function.

        Args:
            function: Function to run
            target: Where the results from the function should be viewed
            controls: Parameters for the Form
            swap: The mechanism to replace the referenced block
            trigger: How the function is called
            cache: Hint that the function output may be cached


        Returns: A Compute block representing the Dynamic behaviour
        """
        if controls is None:
            self.controls = Controls.empty()
        elif isinstance(controls, Controls):
            self.controls = controls
        elif isinstance(controls, t.Mapping):
            self.controls = Controls(**controls)
        else:  # iterable
            self.controls = Controls(*controls)

        self.function = function
        self.cache = cache
        if isinstance(target, BaseBlock):
            if target.name is None:
                raise DPClientError(f"Block {target} must have a name to be used as a target")
            self.target = target.name
        else:
            self.target = str(target)

        # NOTE - using id for the function_id means we can't persist sessions across instances
        # prob ok for now
        self.function_id = f"app.{id(self)}"

        if target in (TargetMode.BELOW, TargetMode.SIDE):
            swap = Swap.INNER
        elif target == TargetMode.SELF:
            swap = Swap.REPLACE
        self.swap = swap

        opt_attribs = {}
        # submit validation
        if trigger == Trigger.SUBMIT:
            opt_attribs.update(submit_label=submit_label)
        elif trigger == Trigger.SCHEDULE:
            opt_attribs.update(timer=timer, immediate=immediate)
        elif trigger in (Trigger.LOAD, Trigger.VISIBLE):
            trigger = Trigger.MOUNT

        # basic attributes
        super().__init__(
            function_id=self.function_id,
            label=label,
            trigger=trigger,
            swap=swap,
            name=name,
            target=target,
            **opt_attribs,
        )


# Compute Helpers
def Form(
    on_submit: FunctionT,
    controls: ControlsT = None,
    target: TargetT = TargetMode.BELOW,
    label: t.Optional[str] = None,
    submit_label: str = "Go",
    cache: bool = False,
) -> Compute:
    """
    Create a form on the page that runs the provided function on submit.

    Args:
        on_submit: Function to run when the form is submitted
        controls: Parameters for the Form
        target (optional): Where the results from the function should be viewed
        label (optional): Description for the Form
        submit_label: Label for the submit button
        cache: Hint that the on_submit function may be cached

    Returns: A Compute block representing the Form
    """
    return Compute(
        function=on_submit, controls=controls, target=target, label=label, submit_label=submit_label, cache=cache
    )


def Dynamic(
    on_load: t.Optional[FunctionT] = None,
    on_timer: t.Optional[FunctionT] = None,
    target: t.Optional[TargetT] = None,
    seconds: int = 30,
    immediate: bool = False,
) -> Compute:
    """
    Create a dynamically updating block that runs the provided function to update the View.

    Args:
        on_load: Function to run when the App is loaded
        on_timer: Function to run on a regular timer
        target (optional): Where the results from the function should be viewed
        seconds (optional, default=30): Number of seconds between running the function specified by on_timer
        immediate (optional, default=False): Run the specified function immediately on page load

    Returns: A Compute block representing the Dynamic behaviour
    """

    if on_load and on_timer:
        raise DPClientError("Can't use both on_load and on_timer simultaneously")
    elif on_load:
        f = on_load
        target = target or TargetMode.SELF
        trigger = Trigger.LOAD
    elif on_timer:
        f = on_timer
        target = target or TargetMode.BELOW
        trigger = Trigger.SCHEDULE
    else:
        raise DPClientError("Must provide one of on_load or on_timer")

    return Compute(function=f, target=target, timer=seconds, trigger=trigger, immediate=immediate)



from abc import ABC, abstractmethod


class ISchema(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def insert_schema(self, polly_session, repo_key, body: dict):
        pass

    @abstractmethod
    def update_schema(self, polly_session, repo_key, body: dict):
        pass

    @abstractmethod
    def replace_schema(self, polly_session, repo_key, body: dict):
        pass

    @abstractmethod
    def validate_schema(self, polly_session, body: dict):
        pass

    @abstractmethod
    def get_schema(
        self,
        polly_session,
        repo_key,
        schema_level=[],
        source="",
        data_type="",
        return_type="dataframe",
    ):
        pass



# Copyright 2020 StackHut Limited (trading as Datapane)
# SPDX-License-Identifier: Apache-2.0

import argparse
import os
import pwd
import sys
import traceback
from pathlib import Path
from typing import Dict, Iterable, List, Mapping, Optional, TextIO, Tuple

from datapane import __version__
from datapane import cloud_api as api
from datapane.client.utils import _setup_dp_logging
from datapane.common import log
from datapane.common.versioning import is_version_compatible

from .config import RunnerConfig, decode
from .exceptions import ModelRunError
from .typedefs import ErrorResult, RunResult


def setup_api(dp_token: str, dp_host: str, debug: bool = False, logs: TextIO = None):
    """Init the Datapane API for automated use"""
    # login, ping, and create the default env file for CMD usage
    api.login(token=dp_token, server=dp_host, cli_login=False)
    # setup input and config, logging, login, etc.
    verbosity = 2 if debug else 0
    _setup_dp_logging(verbosity=verbosity, logs_stream=logs)
    log.debug("Running DP on DP")


def run_api(run_config: RunnerConfig) -> RunResult:
    """Bootstrap the recursive calls into run"""
    app = api.LegacyApp.by_id(run_config.app_id)
    # is the app compatible with the client runner/api
    if not is_version_compatible(__version__, app.api_version, raise_exception=False):
        log.warning(
            f"App developed for an older version of Datapane ({app.api_version}) - "
            + "this run may fail, please update."
        )

    # TODO - we should pull param defaults from app and add in the call
    app.call(run_config.env, **run_config.format())

    # create the RunResult
    script_result = str(api.Result.get()) if api.Result.exists() else None
    report_id = None
    try:
        report = api._report.pop()
        log.debug(f"Returning report id {report.id}")
        report_id = report.id
    except IndexError:
        log.debug("User app didn't generate report - perhaps result / action only")

    return RunResult(report_id=report_id, script_result=script_result)


# ensure we copy PYTHONPATH across
ENV_ALLOWLIST = ("LANG", "PATH", "HOME", "PYTHON_VERSION", "PYTHONPATH")


def make_env(input_env: Mapping[str, str], allowed: Iterable[str] = ENV_ALLOWLIST) -> Dict[str, str]:
    return {k: input_env[k] for k in allowed if k in input_env}


def drop_privileges(out_dir: Path, uid_name: str = "nobody", gid_name: str = "nogroup") -> None:
    """Drop to non-root user if required. TODO - do we still need this??"""
    import grp
    import subprocess

    # from: https://stackoverflow.com/questions/2699907/dropping-root-permissions-in-python
    if os.getuid() != 0:  # type: ignore [attr-defined]
        return

    try:
        # Get the uid/gid from the name
        running_uid = pwd.getpwnam(uid_name).pw_uid  # type: ignore [attr-defined]
        running_gid = grp.getgrnam(gid_name).gr_gid  # type: ignore [attr-defined]

        # chown the output dir
        subprocess.run(["chown", "-R", f"{uid_name}:{gid_name}", str(out_dir)], check=True)

        # Remove group privileges
        os.setgroups([])  # type: ignore [attr-defined]

        # Try setting the new uid/gid
        os.setgid(running_gid)  # type: ignore [attr-defined]
        os.setuid(running_uid)  # type: ignore [attr-defined]

        # Ensure a very conservative umask
        os.umask(0o77)
    except Exception:
        log.warning("Error dropping privileges to non-root user, continuing")


def preexec(out_dir: Path):
    # Currently disabled as unsure how will react with 3rd-party libs, e.g. boto, etc.
    # drop_privileges(out_dir)
    pass


def parse_args(argv: Optional[List[str]] = None) -> Tuple[argparse.Namespace, List[str]]:
    # TODO - min version param
    parser = argparse.ArgumentParser()
    parser.add_argument("--debug", action="store_true", default=False, help="Enable debug output")
    parser.add_argument("--config", required=True)
    parser.add_argument("--out-dir", type=Path, required=True)
    parser.add_argument("--dp-host", required=True)
    parser.add_argument("--dp-token", required=True)
    parser.add_argument("--server-version")
    return parser.parse_known_args(argv)


def main():
    args, unknown_args = parse_args()

    out_dir: Path = args.out_dir
    preexec(out_dir)
    retcode: int

    with (out_dir / "logs.txt").open("w") as stream_logs, (out_dir / "results.json").open("w") as results_stream:
        try:
            # check cli compatible with server - should never really fail
            if args.server_version:
                is_version_compatible(args.server_version, __version__)

            # read the config files
            # NOTE - we could supply config also by env-var or already write k8s volume in the future
            run_config = decode(args.config, compressed=True)
            setup_api(args.dp_token, args.dp_host, args.debug, stream_logs)
            res = run_api(run_config)
        except ModelRunError as e:
            ErrorResult(error=e.error, error_detail=e.details).to_json(results_stream)
            retcode = 103  # NOTE - all user errors return 103
        except Exception:
            log.exception("Unhandled Exception in Model Runner")
            # we could send traceback as details, but not useful to end user
            ErrorResult(error="Unhandled Exception", debug=traceback.format_exc()).to_json(results_stream)
            retcode = 104
        else:
            res.to_json(results_stream)
            retcode = 0

    sys.exit(retcode)


if __name__ == "__main__":
    main()
